from abc import ABC, abstractmethod

class MultiObjectiveSampler:

    def set_priority_graph(self, graph):
        self.priority_graph = graph

    """
    Update distribution from list of objectives.

    Arguments:
        sample: the sample as outputted by ScenicSampler.nextSample()
        info: Contains any necessary information to perform the update (such as bucket indices, etc.)
        rho: A NumPy array with the real objective values [f1, f2, f3, ..., fn]
    """
    @abstractmethod
    def update_dist_from_multi(self, sample, info, rho):
        pass



from .domain_sampler import SamplingError, SplitSampler, TerminationException
from .feature_sampler import FeatureSampler, LateFeatureSampler
from .halton import HaltonSampler
from .cross_entropy import (CrossEntropySampler, ContinuousCrossEntropySampler,
    DiscreteCrossEntropySampler)
from .random_sampler import RandomSampler
from .rejection import RejectionSampler, RejectionException
from .bayesian_optimization import BayesOptSampler
from .simulated_annealing import SimulatedAnnealingSampler
from .scenic_sampler import ScenicSampler
from .grid_sampler import GridSampler
from .dist_BO import DistBayesOptSampler


"""Halton sampler"""

from verifai.samplers.domain_sampler import BoxSampler
import itertools
from math import floor

def generate_primes():
    D = {}
    yield 2
    for q in itertools.islice(itertools.count(3), 0, None, 2):
        p = D.pop(q, None)
        if p is None:
            D[q * q] = q
            yield q
        else:
            x = q + 2 * p
            while x in D:
                x += 2 * p
            D[x] = p

def halton_sequence(index, base):
    f,r = 1, 0
    while index > 0:
        f = f/base
        r += f*(index%base)
        index = floor(index/base)
    return r

class HaltonSampler(BoxSampler):
    """Samples along a quasi-random Halton sequence"""

    def __init__(self, domain, halton_params):
        super().__init__(domain)
        self.sample_index = halton_params.sample_index
        primes = generate_primes()
        for _ in range(halton_params.bases_skipped):
            next(primes)
        self.prime_bases = [next(primes) for _ in range(self.dimension)]

    def getVector(self):
        self.sample_index += 1
        return tuple(halton_sequence(index=self.sample_index, base=p)
                     for p in self.prime_bases), None



"""Rejection sampling"""

from verifai.utils.utils import RejectionException
from verifai.samplers.domain_sampler import ConstrainedSampler, SamplingError

class RejectionSampler(ConstrainedSampler):
    """Enforces a spec over some other sampler by rejection.

    RejectionSampler(RandomSampler(domain), spec)
    """

    def __init__(self, sampler, spec=None, rejectionRho=None, maxRejections=1000):
        super().__init__(sampler.domain, spec)
        self.sampler = sampler
        self.specification = spec
        self.rejectionRho = None
        self.maxRejections = maxRejections

    def getSample(self):
        reject = True
        samples = 0
        while reject:
            if samples >= self.maxRejections:
                raise SamplingError(
                    f'exceeded RejectionSampler limit of {samples} rejections')
            samples += 1
            try:
                sample, info = self.sampler.getSample()
                if self.specification is not None:
                    reject = not self.specification.isSatisfiedBy(sample)
                else:
                    reject = False
                if reject:
                    self.sampler.update(sample, info, self.rejectionRho)
            except RejectionException:
                reject = True
        return sample, info

    def update(self, sample, info, rho):
        self.sampler.update(sample, info, rho)

    def __repr__(self):
        return (f'RejectionSampler({self.sampler}, spec={self.specification}, '
                f'maxRejections={self.maxRejections})')



import carla
import sys

from agents.navigation.agent import *
from agents.navigation.controller import VehiclePIDController
from agents.tools.misc import distance_vehicle, draw_waypoints

import numpy as np

'''This agent doesn't move.'''
class BrakeAgent(Agent):

    def __init__(self, vehicle, opt_dict=None):
        super(BrakeAgent, self).__init__(vehicle)

    def run_step(self):
        control = carla.VehicleControl()
        control.throttle = 0.0
        control.brake = 1.0
        control.hand_brake = True
        return control



# This files defiens the error table as a panda object
import pandas as pd
import numpy as np
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from dotmap import DotMap
from collections import defaultdict
from kmodes.kmodes import KModes

class error_table():
    def __init__(self, space=None, table=None, column_type = None):
        assert space is not None or table is not None
        if space is not None:
            self.space= space
            self.column_names = []
            self.column_type = {}
            for i in range(space.fixedFlattenedDimension):
                self.column_names.append(space.meaningOfFlatCoordinate(i))
                self.column_type[space.meaningOfFlatCoordinate(i)] = \
                    space.coordinateIsNumerical(i)
            self.column_names.append("rho")
            self.column_type["rho"] = True # Set to numerical by default. Can be updated later.
            self.table = pd.DataFrame(columns=self.column_names)
            self.ignore_locs = []
        else:
            self.table = table
            self.column_names = table.columns
            if column_type is None:
                self.column_type = {col:True for col in self.column_names}
            else:
                self.column_type = column_type
            self.ignore_locs = []



    def update_column_names(self, column_names):
        assert len(self.table.columns) == len(column_names)
        self.table.columns = column_names
        self.column_names = column_names

    def update_error_table(self, sample, rho):
        sample = self.space.flatten(sample, fixedDimension=True)
        sample_dict = {}
        for k, v in zip(self.table.columns, list(sample)):
            if np.any(np.array(sample) == None):
                locs = np.where(np.array(sample) == None)
                self.ignore_locs = self.ignore_locs + list(locs[0])
            sample_dict[k] = float(v) if self.column_type[k] and v is not None else v
        if isinstance(rho, (list, tuple)):
            for i,r in enumerate(rho[:-1]):
                if "rho_" + str(i) not in self.column_names:
                    self.column_names.append("rho_"+str(i))
                    if isinstance(r, bool):
                        self.column_type["rho_" + str(i)] = False
                    else:
                        self.column_type["rho_" + str(i)] = True
                sample_dict["rho_"+str(i)]  = r
            sample_dict["rho"] = rho[-1]
            if isinstance(rho[-1], bool) and self.column_type["rho"]:
                self.column_type["rho"] = False
        else:
            sample_dict["rho"] = rho
            if isinstance(rho, bool) and self.column_type["rho"]:
                self.column_type["rho"] = False
        self.ignore_locs = list(set(tuple(self.ignore_locs)))
        new_row = pd.DataFrame(sample_dict, index=[0])
        self.table = pd.concat([self.table, new_row], ignore_index=True)


    def get_column_by_index(self, index):
        if isinstance(index, int):
            index = list([index])
        if len(index) < 1:
            print("No indices provided: returning all samples")
        elif max(index) >= len(self.table.columns):
            for i in index:
                if i >= len(self.table.columns):
                    index.remove(i)
            print("Tried to access index not in error table")
        if len(self.table) > 0:
            names_index = self.table.columns[index]
            return self.table[names_index]
        else:
            print("No entries in error table yet")
        return None

    def get_column_by_name(self, column_names):
        index = []
        if isinstance(column_names, str):
            if column_names in self.table.columns:
                index.append(column_names)
        else:
            for s in column_names:
                if s in self.table.columns:
                    index.append(s)
        return self.table[index]

    def get_samples_by_index(self, index):
        if isinstance(index, int):
            index = list([index])
        if max(index) >= len(self.table):
            print("Trying to access samples not in the table")
            for i in index:
                if i >= len(self.table):
                    index.remove(i)
        return self.table.iloc[index]

    def split_table(self, column_names=None):
        if column_names is None:
            column_names = self.column_names
        numerical, categorical = [], []
        for c in column_names:
            if self.column_type[c]:
                numerical.append(c)
            else:
                categorical.append(c)
        return self.get_column_by_name(numerical), self.get_column_by_name(categorical)

    def get_random_samples(self, count=5):
        if count > len(self.table):
            return list(range(len(self.table)))
        else:
            sample_ids = set()
            while len(sample_ids) < count:
                i = np.random.randint(len(self.table))
                sample_ids.add(i)
            return list(sample_ids)

    def build_normalized(self, column_names=None):
        if len(self.table) < 1:
            return pd.DataFrame(), pd.DataFrame()
        if column_names is None:
            column_names = self.column_names

        numerical, categorical = self.split_table(column_names=column_names)


        if len(categorical.columns) + len(numerical.columns) == 0:
            return pd.DataFrame(), pd.DataFrame()

        # Normalize tables (only for numerical table)
        stats = numerical.describe()

        normalized_dict = {r: (numerical[r] - stats[r]['min']) / (stats[r]['max'] - stats[r]['min'])
                         for r in numerical.columns}
        normalized_table = pd.DataFrame(normalized_dict)

        return normalized_table, categorical, \
               np.array([stats[r]['min'] for r in numerical.columns]),\
               np.array([stats[r]['max'] for r in numerical.columns])


    def build_standardized(self, column_names=None):
        if len(self.table) < 1:
            return pd.DataFrame(), pd.DataFrame()
        if column_names is None:
            column_names = self.column_names

        numerical, categorical = self.split_table(column_names=column_names)

        if len(categorical.columns) + len(numerical.columns) == 0:
            return pd.DataFrame(), pd.DataFrame()

        # Normalize tables (only for numerical table)
        stats = numerical.describe()

        standardized_dict = {r: (numerical[r] - stats[r]['mean']) / stats[r]['std']
                             for r in numerical.columns}
        standardized_table = pd.DataFrame(standardized_dict)

        return standardized_table, categorical, \
               np.array([stats[r]['mean'] for r in numerical.columns]),\
               np.array([stats[r]['std'] for r in numerical.columns])

    def dist_element(self, numerical, point_n):

        d=np.zeros(len(self.table))
        if len(numerical.columns) > 0:
            d = np.linalg.norm(numerical.values- point_n, axis=1)
        return d


    def k_clusters(self, column_names=None, k=None):
        if k is None or k >= len(self.table):
            return np.array(range(len(self.table)))

        if len(self.table) <=0 :
            return np.array(range(len(self.table)))

        numerical, categorical, min_elems, max_elems = self.build_normalized(column_names=column_names)
        range_elems = max_elems- min_elems
        result = defaultdict(dict)


        if len(categorical.columns) > 0:
            X = np.array(categorical.values)
            Y = np.array(numerical.values)
            kmodes = KModes(n_clusters=k, init='Huang', n_init=5, verbose=1)
            kmodes.fit_predict(X)
            centers_cat = kmodes.cluster_centroids_
            labels_cat = kmodes.labels_
            result['categorical'] = {'clusters':centers_cat, 'labels':labels_cat}
            for j in range(k):
                pos = Y[np.where(labels_cat == j)]
                kmeans = KMeans(n_clusters=min(len(pos), k), random_state=0).fit(pos)
                centers = kmeans.cluster_centers_
                labels = kmeans.labels_
                for label, center in enumerate(centers):
                    center = (center * range_elems) + min_elems
                    result[j][label] = center
                result[j]['labels'] = labels

        elif len(numerical.columns) > 0:
            X = np.array(numerical.values)
            kmeans = KMeans(n_clusters=k, random_state=0).fit(X)
            centers = kmeans.cluster_centers_
            labels = kmeans.labels_
            for label, center in enumerate(centers):
                center = (center*range_elems) + min_elems
                result['clusters'][label] = center
            result['labels'] = labels
        return result


    def k_closest_samples(self, column_names=None, k = None, dist_type=True):
        # dist_type is True for using normalized, False for standardized
        if k is None or k >= len(self.table):
             return np.array(range(len(self.table)))

        if dist_type:
            numerical, categorical, _, _ = self.build_normalized(column_names=column_names)
        else:
            numerical, categorical, _, _ = self.build_standardized(column_names=column_names)

        # Norm distance between table rows
        d_rows = np.zeros((len(self.table), len(self.table)))
        for i in range(len(self.table)):
            d_rows[i] = self.dist_element(numerical, numerical.values[i])

        # Now the row associated with the min sum is the largest set of correlated elements
        sum_rows = []
        correlated_rows = []
        for r in d_rows:
            ks = np.argpartition(r, k)[:k]
            sum_rows.append(sum(r[ks]))
            correlated_rows.append(ks)

        return correlated_rows[np.array(sum_rows).argmin()]

    def pca_analysis(self, column_names=None, n_components= 1):
        # Returns the direction of the principal component among the samples
        if len(self.table) < 1:
            return

        if column_names is None:
            column_names = self.column_names

        numerical, _ = self.split_table(column_names=column_names)


        # Do PCA analysis only on the numerical columns
        if len(numerical.columns) == 0:
            return
        pca_columns = []
        for c in numerical.columns:
            if c in self.table.columns and c not in self.table.columns[self.ignore_locs]:
                pca_columns.append(c)

        table = self.get_column_by_name(pca_columns)

        # PCA components
        if n_components > min(len(table), len(table.columns)):
            n_components = min(len(table), len(table.columns))

        pca = PCA(n_components=n_components)
        pca.fit(table)

        return {'columns':pca_columns, 'pivot': table.mean().values, 'directions': pca.components_}

    def analyze(self, analysis_params=None):
        analysis_data = DotMap()
        if analysis_params is None or ('pca' in analysis_params and analysis_params.pca) or 'pca' not in analysis_params:
            if analysis_params is not None and 'pca_params' in analysis_params:
                columns = analysis_params.pca_params.columns \
                    if 'columns' in analysis_params.pca_params else None
                n_components = analysis_params.pca_params.n_components \
                    if 'n_components' in analysis_params.pca_params else 1
            else:
                columns, n_components = None, 1
            analysis_data.pca = self.pca_analysis(column_names=columns, n_components=n_components)

        if analysis_params is None or ('k_closest' in analysis_params and analysis_params.k_closest) or 'k_closest' not in analysis_params:
            if analysis_params is not None and 'k_closest_params' in analysis_params:
                columns = analysis_params.k_closest_params.columns \
                    if 'columns' in analysis_params.k_closest_params else None
                k = analysis_params.k_closest_params.k \
                    if 'k' in analysis_params.k_closest_params else None
            else:
                columns, k = None, None
            analysis_data.k_closest = self.k_closest_samples(column_names=columns, k=k)


        if analysis_params is None or ('random' in analysis_params and analysis_params.random) or 'random' not in analysis_params:
            if analysis_params is not None and 'random_params' in analysis_params:
                count = analysis_params.random_params.count \
                    if 'count' in analysis_params.random_params else 5
            else:
                count = 5
            analysis_data.random = self.get_random_samples(count=count)

        if analysis_params is None or ('k_clusters' in analysis_params and analysis_params.k_clusters) or 'k_clusters' not in analysis_params:
            if analysis_params is not None and 'k_clusters_params' in analysis_params:
                columns = analysis_params.k_clusters_params.columns \
                    if 'columns' in analysis_params.k_clusters_params else None
                k = analysis_params.k_clusters_params.k \
                    if 'k' in analysis_params.k_clusters_params else None
            else:
                columns, k = None, None
            analysis_data.k_clusters = self.k_clusters(column_names=columns, k=k)
        return analysis_data








from enum import IntEnum
import random
import numpy as np

import carla
from agents.navigation.agent import *

from pid_follow_controller import *

class LaneChange(IntEnum):
    LEFT = -1
    RIGHT = 1

'''Child of PIDAgent that also overtakes any blocking vehicles.'''
class OvertakeAgent(PIDAgent):

    def __init__(self, vehicle, opt_dict):
        super().__init__(vehicle, opt_dict)

        self.location_pid_dict = {
            'K_P': 1.0,
            'K_D': 0.05,
            'K_I': 1,
            'dt': 0.05
        }
        # Distance to maintain from other vehicles.
        self.clear_dist = 10.0

        if opt_dict:
            if 'target_speed' in opt_dict:
                self.location_pid_dict['dt'] = 1.0 / self.target_speed
            if 'clear_dist' in opt_dict:
                self.clear_dist = opt_dict['clear_dist']
            if 'location_pid_dict' in opt_dict:
                self.location_pid_dict = opt_dict['location_pid_dict']

        self.controller = PIDFollowController(
            vehicle,
            clear_dist=self.clear_dist,
            args_lateral=self.lateral_pid_dict,
            args_longitudinal=self.longitudinal_pid_dict,
            args_location=self.location_pid_dict)
        # Magnitude of lane_state is how far vehicle is from its
        # desired lane. Negative/positive means to the left/right
        # of desired lane, respectively.
        self.lane_state = 0
        self.is_changing_lane = False


    def get_lane_change_w(self, cur_w, lane_change):
        # Return waypoint corresponding to LANE_CHANGE (either LEFT or RIGHT)
        # from waypoint CUR_W. If lane change illegal or unsafe, returns None.
        next_w = None
        lane_safe = lambda w: self.lane_clear(w.transform.location)[0] and \
            self.lane_clear(w.transform.location,
                            min_dist=10.0*self.clear_dist,
                            forward=w.transform.get_forward_vector())[0]
        if lane_change is LaneChange.LEFT and cur_w.lane_change & carla.LaneChange.Left:
            if lane_safe(cur_w.get_left_lane()):
                next_w = cur_w.get_left_lane()
        elif cur_w.lane_change & carla.LaneChange.Right:
            if lane_safe(cur_w.get_right_lane()):
                next_w = cur_w.get_right_lane()
        return next_w


    def change_lane(self, lane_change=None):
        ''' By default, picks left or right at random. If no possible lane change,
        does nothing.'''
        if lane_change:
            potential_change = [lane_change]
        else:
            potential_change = [LaneChange.LEFT, LaneChange.RIGHT]
        potential_w = []
        cur_w = self.waypoints[0]
        for change in potential_change:
            next_w = self.get_lane_change_w(cur_w, change)
            if next_w:
                potential_w.append((change, next_w))
        if potential_w:
            lane_change, next_w = random.choice(potential_w)
            self.is_changing_lane = True
            self.lane_state += lane_change
            self.waypoints = [cur_w, next_w]


    def lane_clear(self, location, min_dist=None, forward=None):
        '''
        Check that the lane LOCATION is in is clear of vehicles.
        Lane is clear if no vehicles within MAX_DIST away.
        If not clear, return tuple of False and blocking vehicle.
        If FORWARD, only check vehicles along direction of FORWARD.
        '''
        def norm(vec):
            return np.sqrt(vec.x ** 2 + vec.y ** 2 + vec.z ** 2)

        def norm_dot(a, b):
            a /= norm(a)
            b /= norm(b)
            dot = a.x * b.x + a.y * b.y + a.z * b.z
            return dot
        
        if not min_dist:
            min_dist = self.clear_dist
            
        lane_id = self._map.get_waypoint(location).lane_id
        vehicles = self._world.get_actors().filter('*vehicle*')
        for v in vehicles:
            # Check if v is self.
            if v.id == self._vehicle.id:
                continue

            # Check if v is on same lane as self.
            v_loc = v.get_location()
            v_w = self._map.get_waypoint(v_loc)
            if lane_id != v_w.lane_id:
                continue

            if forward and norm_dot(forward, v_loc - location) < 0.0:
                continue

            if v_loc.distance(location) < min_dist:
                return (False, v)
        return (True, None)
        

    def run_step(self):
        self_loc = self._vehicle.get_location()
        self_forward = self._vehicle.get_transform().get_forward_vector()
        is_clear, blocker = self.lane_clear(self_loc,
                                            min_dist=2.0*self.clear_dist,
                                            forward=self_forward)
        super().run_step()
        cur_w = self.waypoints[0]
        if self.lane_state != 0:
            speed = self.target_speed * 1.5
        else:
            speed = self.target_speed

        if not is_clear:
            if not self.is_changing_lane:
                self.change_lane()
                    
            return self.controller.run_step(speed,
                                            self.waypoints[0],
                                            blocker.get_location())
        else:
            self.is_changing_lane = False
            if self.lane_state != 0:
                lane_change = LaneChange(-np.sign(self.lane_state))
                self.change_lane(lane_change=lane_change)
                
            return self.controller.run_step(speed,
                                            self.waypoints[0])





"""Samplers generating points in a feature space, possibly subject to
specifications.
"""

import random

import dill
from dotmap import DotMap
import numpy as np

from verifai.features import FilteredDomain
from verifai.samplers.domain_sampler import SplitSampler, TerminationException
from verifai.samplers.rejection import RejectionSampler
from verifai.samplers.halton import HaltonSampler
from verifai.samplers.cross_entropy import CrossEntropySampler
from verifai.samplers.random_sampler import RandomSampler
from verifai.samplers.multi_armed_bandit import MultiArmedBanditSampler
from verifai.samplers.eg_sampler import EpsilonGreedySampler
from verifai.samplers.bayesian_optimization import BayesOptSampler
from verifai.samplers.simulated_annealing import SimulatedAnnealingSampler
from verifai.samplers.grid_sampler import GridSampler

### Samplers defined over FeatureSpaces

class FeatureSampler:
    """Abstract class for samplers over FeatureSpaces."""

    def __init__(self, space):
        self.space = space
        self.last_sample = None

    @classmethod
    def samplerFor(cls, space):
        """Convenience function choosing a default sampler for a space."""
        return cls.randomSamplerFor(space)

    @staticmethod
    def randomSamplerFor(space):
        """Creates a random sampler for a given space"""
        return LateFeatureSampler(space, RandomSampler, RandomSampler)

    @staticmethod
    def haltonSamplerFor(space, halton_params=None):
        """Creates a Halton sampler for a given space.

        Uses random sampling for lengths of feature lists and any
        Domains that are not continous and standardizable.
        """
        if halton_params is None:
            halton_params = default_sampler_params('halton')
        def makeDomainSampler(domain):
            return SplitSampler.fromPredicate(
                domain,
                lambda d: d.standardizedDimension > 0,
                lambda domain: HaltonSampler(domain=domain,
                                             halton_params=halton_params),
                makeRandomSampler)
        return LateFeatureSampler(space, RandomSampler, makeDomainSampler)

    @staticmethod
    def crossEntropySamplerFor(space, ce_params=None):
        """Creates a cross-entropy sampler for a given space.

        Uses random sampling for lengths of feature lists and any Domains
        that are not standardizable.
        """
        if ce_params is None:
            ce_params = default_sampler_params('ce')
        return LateFeatureSampler(space, RandomSampler,
            lambda domain: CrossEntropySampler(domain=domain,
                                               ce_params=ce_params))

    @staticmethod
    def epsilonGreedySamplerFor(space, eg_params=None):
        """Creates an epsilon-greedy sampler for a given space.

        Uses random sampling for lengths of feature lists and any Domains
        that are not standardizable.
        """
        if eg_params is None:
            eg_params = default_sampler_params('eg')
        return LateFeatureSampler(space, RandomSampler,
            lambda domain: EpsilonGreedySampler(domain=domain,
                                                eg_params=eg_params))

    @staticmethod
    def multiArmedBanditSamplerFor(space, mab_params=None):
        """Creates a multi-armed bandit sampler for a given space.

        Uses random sampling for lengths of feature lists and any Domains
        that are not standardizable.
        """
        if mab_params is None:
            mab_params = default_sampler_params('mab')
        return LateFeatureSampler(space, RandomSampler,
            lambda domain: MultiArmedBanditSampler(domain=domain,
                                                   mab_params=mab_params))

    @staticmethod
    def gridSamplerFor(space, grid_params=None):
        """Creates a grid sampler for a given space.

        Uses random sampling for lengths of feature lists and any Domains
        that are not standardizable."""

        def makeDomainSampler(domain):
            return SplitSampler.fromPredicate(
                domain,
                lambda d: d.isStandardizable,
                lambda domain: GridSampler(domain=domain,
                                           grid_params=grid_params),
                makeRandomSampler)
        return LateFeatureSampler(space, RandomSampler, makeDomainSampler)

    @staticmethod
    def simulatedAnnealingSamplerFor(space, sa_params=None):
        """Creates a cross-entropy sampler for a given space.

        Uses random sampling for lengths of feature lists and any Domains
        that are not continuous and standardizable.
        """
        if sa_params is None:
            sa_params = default_sampler_params('sa')
        def makeDomainSampler(domain):
            return SplitSampler.fromPredicate(
                domain,
                lambda d: d.standardizedDimension > 0,
                lambda domain: SimulatedAnnealingSampler(domain=domain,
                                                         sa_params=sa_params),
                makeRandomSampler)
        return LateFeatureSampler(space, RandomSampler, makeDomainSampler)

    @staticmethod
    def bayesianOptimizationSamplerFor(space, BO_params=None):
        """Creates a Bayesian Optimization sampler for a given space.

        Uses random sampling for lengths of feature lists and any
        Domains that are not continous and standardizable.
        """
        if BO_params is None:
            BO_params = default_sampler_params('bo')
        def makeDomainSampler(domain):
            return SplitSampler.fromPredicate(
                domain,
                lambda d: d.standardizedDimension > 0,
                lambda domain: BayesOptSampler(domain=domain,
                                               BO_params=BO_params),
                makeRandomSampler)
        return LateFeatureSampler(space, RandomSampler, makeDomainSampler)

    def getSample(self):
        """Generate a sample, along with any sampler-specific info.

        Must return a pair consisting of the sample and arbitrary
        sampler-specific info, which will be passed to the `update`
        method after the sample is evaluated.
        """
        raise NotImplementedError('tried to use abstract FeatureSampler')
    
    def update(self, sample, info, rho):
        """Update the state of the sampler after evaluating a sample."""
        pass

    def nextSample(self, feedback=None):
        """Generate the next sample, given feedback from the last sample.

        This function exists only for backwards compatibility. It has been
        superceded by the `getSample` and `update` APIs.
        """
        if self.last_sample is not None:
            self.update(self.last_sample, self.last_info, feedback)
        self.last_sample, self.last_info = self.getSample()
        return self.last_sample

    def set_graph(self, graph):
        self.scenario.set_graph(graph)

    def saveToFile(self, path):
        with open(path, 'wb') as outfile:
            randState = random.getstate()
            numpyRandState = np.random.get_state()
            allState = (randState, numpyRandState, self)
            dill.dump(allState, outfile)

    @staticmethod
    def restoreFromFile(path):
        with open(path, 'rb') as infile:
            allState = dill.load(infile)
            randState, numpyRandState, sampler = allState
            random.setstate(randState)
            np.random.set_state(numpyRandState)
            return sampler

    def __iter__(self):
        try:
            feedback = None
            while True:
                feedback = yield self.nextSample(feedback)
        except TerminationException:
            return

class LateFeatureSampler(FeatureSampler):
    """FeatureSampler that works by first sampling only lengths of feature
    lists, then sampling from the resulting fixed-dimensional Domain.

    e.g. LateFeatureSampler(space, RandomSampler, HaltonSampler) creates a
    FeatureSampler which picks lengths uniformly at random and applies
    Halton sampling to each fixed-length space.
    """

    def __init__(self, space, makeLengthSampler, makeDomainSampler):
        super().__init__(space)
        lengthDomain, fixedDomains = space.domains
        if lengthDomain is None:    # space has no feature lists
            self.lengthSampler = None
            self.domainSampler = makeDomainSampler(fixedDomains)
        else:
            self.lengthDomain = lengthDomain
            self.lengthSampler = makeLengthSampler(lengthDomain)
            self.domainSamplers = {
                point: makeDomainSampler(domain)
                for point, domain in fixedDomains.items()
            }
            self.lastLength = None

    def getSample(self):
        if self.lengthSampler is None:
            domainPoint, info = self.domainSampler.getSample()
        else:
            length, info1 = self.lengthSampler.getSample()
            self.lastLength = length
            domainPoint, info2 = self.domainSamplers[length].getSample()
            info = (info1, info2)
        return self.space.makePoint(*domainPoint), info
    
    def update(self, sample, info, rho):
        if self.lengthSampler is None:
            self.domainSampler.update(sample, info, rho)
        else:
            self.lengthSampler.update(sample, info[0], rho)
            lengths = []
            for name, feature in self.space.namedFeatures:
                if feature.lengthDomain:
                    lengths.append((len(getattr(sample, name)),))
            lengthPoint = self.lengthDomain.makePoint(*lengths)
            self.domainSamplers[lengthPoint].update(sample, info[1], rho)

### Utilities

def makeRandomSampler(domain):
    """Utility function making a random sampler for a domain."""
    sampler = RandomSampler(domain)
    if domain.requiresRejection:
        sampler = RejectionSampler(sampler)
    return sampler

def default_sampler_params(sampler_type):
    if sampler_type == 'halton':
        return DotMap(sample_index=0, bases_skipped=0)
    elif sampler_type in ('ce', 'eg', 'mab'):
        cont = DotMap(buckets=5, dist=None)
        disc = DotMap(dist=None)
        return DotMap(alpha=0.9, thres=0.0, cont=cont, disc=disc)
    elif sampler_type == 'bo':
        return DotMap(init_num=5)
    return DotMap()



import carla
from agents.navigation.agent import *
from agents.navigation.controller import VehiclePIDController
from agents.tools.misc import distance_vehicle, draw_waypoints

import numpy as np

'''Agent that follows road waypoints (prioritizing a straight
trajectory if multiple options available) using longitudinal
and lateral PID.'''
class PIDAgent(Agent):

    def __init__(self, vehicle, opt_dict=None):
        super(PIDAgent, self).__init__(vehicle)

        # Default agent params:
        self.target_speed = 20.0

        # Default PID params:
        self.lateral_pid_dict = {
            'K_P': 1.0,
            'K_D': 0.01,
            'K_I': 0.6,
            'dt': 0.05
        }
        self.longitudinal_pid_dict = {
            'K_P': 1.0,
            'K_D': 0,
            'K_I': 1,
            'dt': 0.05
        }
        if opt_dict:
            if 'target_speed' in opt_dict:
                self.target_speed = opt_dict['target_speed']
                self.lateral_pid_dict['dt'] = 1.0 / self.target_speed
                self.longitudinal_pid_dict['dt'] = 1.0 / self.target_speed
            if 'radius' in opt_dict:
                self.radius = opt_dict['radius']
            if 'min_dist' in opt_dict:
                self.min_dist = opt_dict['min_dist']
            if 'lateral_pid_dict' in opt_dict:
                self.lateral_pid_dict = opt_dict['lateral_pid_dict']
            if 'longitudinal_pid_dict' in opt_dict:
                self.longitudinal_pid_dict = opt_dict['longitudinal_pid_dict']
        self.controller = VehiclePIDController(vehicle,
                                              args_lateral=self.lateral_pid_dict,
                                              args_longitudinal=self.longitudinal_pid_dict)
        self.waypoints = []
        self.max_waypoints = 200

        self.radius = self.target_speed / 3.6    # Radius at which next waypoint is sampled
        self.min_dist = 0.9 * self.radius  # If min_dist away from waypoint[0], move on to next one.


    def add_next_waypoints(self):
        def d_angle(a, b):
            return abs((a - b + 180) % 360 - 180)

        if not self.waypoints:
            current_w = self._map.get_waypoint(self._vehicle.get_location())
            self.waypoints.append(current_w)
        while len(self.waypoints) < self.max_waypoints:
            last_w = self.waypoints[-1]
            last_heading = last_w.transform.rotation.yaw
            next_w_list = list(last_w.next(self.radius))
            # Go straight if possible.
            if next_w_list:
                next_w  = min(next_w_list,
                              key = lambda w: d_angle(w.transform.rotation.yaw, last_heading))
            else:
                print('No more waypoints.')
                return
            self.waypoints.append(next_w)


    def run_step(self):
        transform = self._vehicle.get_transform()

        if self.waypoints:
            # If too far off course, reset waypoint queue.
            if distance_vehicle(self.waypoints[0], transform) > 5.0 * self.radius:
                self.waypoints = []

        # Get more waypoints.
        if len(self.waypoints) < self.max_waypoints // 2:
            self.add_next_waypoints()

        # If no more waypoints, stop.
        if not self.waypoints:
            print('Ran out of waypoints; stopping.')
            control = carla.VehicleControl()
            control.throttle = 0.0
            return control


        # Remove waypoints we've reached.
        while distance_vehicle(self.waypoints[0], transform) < self.min_dist:
            self.waypoints = self.waypoints[1:]

        # Draw next waypoint
        draw_waypoints(self._vehicle.get_world(),
                           self.waypoints[:1],
                           self._vehicle.get_location().z + 1.0)

        return self.controller.run_step(self.target_speed, self.waypoints[0])




# This file defines a lane
import numpy as np

class straight_lane():
    def __init__(self, p, q, w):
        self.p = np.array(p)
        self.q = np.array(q)
        self.w = w
        self.m = (self.q-self.p)/np.linalg.norm(self.p-self.q)
        self.n = np.array([-self.m[1], self.m[0]])

    def shifted(self, m):
        return straight_lane(self.p+self.n*self.w*m,
                             self.q+self.n*self.w*m, self.w)

    def dist(self, x):
        r = (x[0] - self.p[0]) * self.n[0] + (x[1] - self.p[1]) * self.n[1]
        return np.sqrt(r * r)





from verifai.features.features import *
from verifai.falsifier import mtl_falsifier
from dotmap import DotMap

init_conditions = Struct({
        'x_init': Box([-0.45, 0.45]),
        'cruising_speed': Box([10.0, 20.0]),
        'theta_init': Box([-np.pi/4, np.pi/4])
    })

sample_space = {'init_conditions': init_conditions}

SAMPLERTYPE = 'ce'
MAX_ITERS = 20
PORT = 8888
MAXREQS = 5
BUFSIZE = 4096

specification = ["G(xdeviation)"]

falsifier_params = DotMap()
falsifier_params.n_iters = MAX_ITERS
falsifier_params.verbosity = 1
falsifier_params.compute_error_table = True
falsifier_params.fal_thres = 0.0

server_options = DotMap(port=PORT, bufsize=BUFSIZE, maxreqs=MAXREQS)


sampler_params = DotMap()
sampler_params.init_num = 2
falsifier_params.sampler_params = sampler_params

falsifier = mtl_falsifier(sample_space=sample_space, sampler_type=SAMPLERTYPE,
                             specification=specification, falsifier_params=falsifier_params,
                          server_options=server_options)
falsifier.run_falsifier()
analysis_params = DotMap()
analysis_params.k_closest_params.k = 4
analysis_params.random_params.count = 4
falsifier.analyze_error_table(analysis_params=analysis_params)

print("Falsified Samples")
print(falsifier.error_table.table)

print("Safe Samples")
print(falsifier.safe_table.table)


# To save all samples: uncomment this
# pickle.dump(falsifier.samples, open("generated_samples.pickle", "wb"))





import numpy as np
from scipy import linalg

def extract_AB(speed, dt, wheelbase=2.995):
    # if state is true, doing lane keeping
    A = np.array([[1.0, 0.0, dt*speed*5./18.],[0.0, 1.0, 0.0],[0.0, 0.0, 1.0]])
    B = np.array([[0.0], [0.0], [(dt*speed*5./18.)/wheelbase]])
    return A, B

def discrete_LQR(A, B, Q, R, c):
    nX, nU = np.shape(B)[0], np.shape(B)[1]

    A_new = np.concatenate((A, c[:, np.newaxis]), axis=1)
    A_new = np.concatenate((A_new, np.concatenate((np.zeros((1, nX)), np.ones((1, 1))), axis=1)), axis=0)

    B_new = np.concatenate((B, np.zeros((1, nU))), axis=0)

    Q_new = linalg.block_diag(Q, np.array([[0.]]))

    K_current = np.zeros((nU, nX + 1))
    P_current = Q_new
    flag = True

    while flag:
        M1 = R + B_new.T.dot(P_current).dot(B_new)
        K_new = -1 * np.linalg.inv(M1).dot(B_new.T).dot(P_current).dot(A_new)
        M2 = A_new + B_new.dot(K_new)
        P_new = Q_new + K_new.T.dot(R).dot(K_new) + M2.T.dot(P_current).dot(M2)
        if (np.linalg.norm(K_new - K_current, 2) <= 1e-4):
            flag = False
        else:
            K_current = K_new
            P_current = P_new
    K = K_current
    P = P_current

    K_actual = K[:nU, :nX]
    k_actual = K[:, -1]

    return K_actual, k_actual

def continuous_LQR(speed, Q, R, wheelbase=2.995):
    A= np.matrix([[0, speed], [0, 0]])
    B = np.matrix([[0], [(speed/wheelbase)]])
    V = np.matrix(linalg.solve_continuous_are(A, B, Q, R))
    K = np.matrix(linalg.inv(R) * (B.T * V))
    return K





""" Grid sampling (Only use for low dimensional spaces)"""

import itertools
import numpy as np
from verifai.samplers.domain_sampler import (BoxSampler, DiscreteBoxSampler,
    DomainSampler, SplitSampler, IteratorSampler, TerminationException)
from verifai.samplers.random_sampler import RandomSampler

class GridSampler(IteratorSampler):
    def __init__(self, domain, grid_params=None):
        if grid_params is None:
            grid_params = {}
        self.cont_N = grid_params.get('N', 21)
        repeat = grid_params.get('repeat', False)
        super().__init__(domain, repeat=repeat)

        cont_grid = lambda domain: ContinuousGridSampler(domain=domain,
                                                              N=self.cont_N)
        disc_grid = lambda domain: DiscreteGridSampler(domain=domain)

        partition = (
            (lambda d: d.standardizedDimension >= 0, cont_grid),
            (lambda d: d.standardizedIntervals, disc_grid)
        )
        self.split_sampler = SplitSampler.fromPartition(domain, partition)

        self.cont_sampler, self.disc_sampler = \
            self.split_sampler.samplersForPredicates

    def __iter__(self):
        while True:
            for subpoints in itertools.product(*self.split_sampler.samplers):
                yield self.domain.rejoinPoints(*subpoints)
            if not self.repeat:
                return

class ContinuousGridSampler(BoxSampler):
    def __init__(self, domain, N):
        super().__init__(domain)

        D = self.domain.standardizedDimension
        if isinstance(N, int):
            self.N = np.ones(D) * N
        else:
            if len(N) != D:
                raise RuntimeError(
                    f'grid specification for continuous space of dimension {D}'
                    f' has wrong length {len(N)}'
                )
            if not all(isinstance(k, int) and k > 0 for k in N):
                raise RuntimeError(
                    f'grid specification {N} must consist of positive integers'
                )
            self.N = N

        self.iters = 0
        self.max_iters = np.prod(self.N)

    def getVector(self):
        sample_vec = []
        t = self.iters
        if t == self.max_iters:
            self.iters = 0
            raise TerminationException('finished continuous grid sampling')

        for k in self.N:
            sample_vec.append(1./(k-1)*(t%k))
            t = t//k
        self.iters+=1
        return tuple(sample_vec), None

class DiscreteGridSampler(DiscreteBoxSampler):
    def __init__(self, domain):
        super().__init__(domain)
        self.start_N = 1
        self.iters = 0
        self.max_iters = 1
        for (left, right) in self.domain.standardizedIntervals:
            self.max_iters *= (right - left) + 1

    def getVector(self):
        if self.iters == self.max_iters:
            self.iters = 0
            raise TerminationException('finished discrete grid sampling')
        sample_vec = []
        t = self.iters
        for (left, right) in self.domain.standardizedIntervals:
            sample_vec.append(left + t%(right-left+1))
            t = t//(right-left +1)
        self.iters+=1
        return tuple(sample_vec), None



import functools, argparse, select
import struct, array
import time
import sys
import math
import socket

import numpy as np
import pandas as pd
import yaml
from dotmap import DotMap

try:
    from xpc import XPlaneConnect
except ImportError as e:
    raise RuntimeError('the X-Plane interface requires XPlaneConnect') from e

import verifai, verifai.server, verifai.falsifier, verifai.samplers
import verifai.simulators.xplane.utils.controller as simple_controller
from verifai.simulators.xplane.utils.geometry import (euclidean_dist, quaternion_for,
    initial_bearing, cross_track_distance, compute_heading_error)

position_dref = ["sim/flightmodel/position/local_x",
                 "sim/flightmodel/position/local_y",
                 "sim/flightmodel/position/local_z"]

orientation_dref = ["sim/flightmodel/position/theta",
                    "sim/flightmodel/position/phi",
                    "sim/flightmodel/position/q"]

velocity_drefs = [      # velocity and acceleration drefs to null between runs
    "sim/flightmodel/position/local_vx",
    "sim/flightmodel/position/local_vy",
    "sim/flightmodel/position/local_vz",
    "sim/flightmodel/position/local_ax",
    "sim/flightmodel/position/local_ay",
    "sim/flightmodel/position/local_az",
    "sim/flightmodel/position/P",
    "sim/flightmodel/position/Q",
    "sim/flightmodel/position/R",
]

fuel_dref = "sim/flightmodel/weight/m_fuel"

class XPlaneServer(verifai.server.Server):
    def __init__(self, sampling_data, monitor, options):
        super().__init__(sampling_data, monitor, options)

        self.runway_heading = options.runway.heading
        self.rh_sin, self.rh_cos = math.sin(self.runway_heading), math.cos(self.runway_heading)
        self.start_lat, self.start_lon = options.runway.start_lat, options.runway.start_lon
        self.end_lat, self.end_lon = options.runway.end_lat, options.runway.end_lon
        self.desired_heading = initial_bearing(self.start_lat, self.start_lon,
                                               self.end_lat, self.end_lon)
        self.origin_x, self.origin_z = options.runway.origin_x, options.runway.origin_z

        path = pd.read_csv(options.runway.elevations)
        self.x, self.y, self.z = path['x'].values, path['y'].values, path['z'].values
        assert len(self.x) == len(self.y), "latitude and longitude values do not align"

        self.controller = options.get('controller')
        self.frametime = 1.0 / options.get('framerate', 10.0)
        self.timeout = options.get('timeout', 8000)
        self.predicates = options.get('predicates', {})
        self.verbosity = options.get('verbosity', 0)

        video = options.get('video')
        if video is not None:
            region = video['region']
            size = video.get('size')
            if size is not None:
                size = (size['width'], size['height'])
            import utils.images as im
            self.grab_image = lambda: im.grab_image(region, resizeTo=size)
        else:
            self.grab_image = None

        try:
            self.xpcserver = XPlaneConnect(timeout=self.timeout)
            input('Please start a new flight and then press ENTER.')
            self.initFuel = list(self.xpcserver.getDREF(fuel_dref))
            theta, phi, q = self.xpcserver.getDREFs(orientation_dref)
            self.initTheta, self.initPhi = math.radians(theta[0]), math.radians(phi[0])
            x, y, z = self.xpcserver.getDREFs(position_dref)
            self.initPosition = (x[0], y[0], z[0])
        except socket.timeout as e:
            raise RuntimeError('unable to connect to X-Plane') from e

    def evaluate_sample(self, sample):
        simulation_data = self.run_simulation(sample)
        value = 0 if self.monitor is None else self.monitor.evaluate(simulation_data)
        return value

    def run_simulation(self, sample):
        # Get runway endpoints
        start_lat, start_lon = self.start_lat, self.start_lon
        end_lat, end_lon = self.end_lat, self.end_lon

        # Extract test parameters from sample
        if isinstance(self.sampler, verifai.ScenicSampler):
            # special case for parameters whose names aren't valid identifiers
            # TODO generalize this mechanism to other sampler types?
            params = self.sampler.paramDictForSample(sample)
        else:
            params = sample.params._asdict()
        simulation_time = params.pop('simulation_length', 30)
        setup_time = params.pop('setup_time', 2)
        # anything left in params we assume to be a dref
        custom_drefs = params
        # get plane position and heading
        plane = sample.objects[0]
        heading = plane.heading
        plane_x_sc, plane_z_sc = plane.position

        # Compute initial plane XZ position and heading in X-Plane local coordinates
        start_heading = self.runway_heading - heading
        plane_x = self.origin_x + (self.rh_cos * plane_x_sc) + (self.rh_sin * plane_z_sc)
        plane_z = self.origin_z - (self.rh_cos * plane_z_sc) + (self.rh_sin * plane_x_sc)

        # Find correct Y position for plane
        x_y = (plane_x, plane_z)
        euclid_dists = np.array([euclidean_dist(x_y, (self.x[i], self.z[i]))
                                for i in range(len(self.x))])
        best_elev = np.argmin(euclid_dists)
        plane_y = self.y[best_elev]

        # Reset plane and apply brake
        self.xpcserver.sendDREFs(velocity_drefs, [0]*len(velocity_drefs))
        self.xpcserver.sendCTRL([0, 0, 0, 0])
        self.xpcserver.sendCOMM("sim/operation/fix_all_systems")
        self.xpcserver.sendDREF(fuel_dref, self.initFuel)
        self.xpcserver.sendDREF("sim/flightmodel/controls/parkbrake", 1)

        # Set new plane position
        self.xpcserver.sendDREFs(position_dref, [plane_x, plane_y, plane_z])
        # Set new plane orientation
        quaternion = quaternion_for(self.initTheta, self.initPhi, start_heading)
        self.xpcserver.sendDREF("sim/flightmodel/position/q", quaternion)
        # Set any other specified drefs
        self.xpcserver.sendDREFs(list(custom_drefs.keys()), list(custom_drefs.values()))

        # Wait for weather, etc. to stabilize, then release brake
        time.sleep(setup_time)
        self.xpcserver.sendCOMM("sim/operation/fix_all_systems")
        time.sleep(0.1)
        self.xpcserver.sendDREF("sim/flightmodel/controls/parkbrake", 0)

        # Execute a run
        if self.verbosity >= 1:
            print('Starting run...')
        start = time.time()
        current = start
        lats, lons, psis, ctes, hes, times, images = [], [], [], [], [], [], []
        while current - start < simulation_time:
            times.append(current - start)
            # Get current plane state
            # Use modified getPOSI to get lat/lon in double precision
            lat, lon, _, _, _, psi, _ = self.xpcserver.getPOSI()
            lats.append(lat); lons.append(lon); psis.append(psi)
            # Compute cross-track and heading errors
            cte = cross_track_distance(start_lat, start_lon, end_lat, end_lon, lat, lon)
            heading_err = compute_heading_error(self.desired_heading, psi)
            ctes.append(cte); hes.append(heading_err)
            # Run controller for one step, if desired
            if self.controller is not None:
                self.controller(self.xpcserver, lat, lon, psi, cte, heading_err)
            # Save screenshot for videos
            if self.grab_image is not None:
                images.append(self.grab_image())
            if self.verbosity >= 2:
                print(f'cte: {cte}; heading_err: {heading_err}')

            # Limit framerate (for fast controllers)
            wait_time = self.frametime - (time.time() - current)
            if wait_time > 0:
                time.sleep(wait_time)

            current = time.time()
        self.images = images

        # Do some simple checks to see if the plane has gotten stuck
        thresh = 0.000001
        end_point_check, mid_point_check = True, True
        if abs(lats[0] - lats[-1]) < thresh and abs(lons[0] - lons[-1]) < thresh:
            end_point_check = False
        num_lats = len(lats)
        if abs(lats[0] - lats[num_lats//2]) < thresh and abs(lons[0] - lons[num_lats//2]) < thresh:
            mid_point_check = False
        if not (mid_point_check or end_point_check):
            raise RuntimeError('Plane appears to have gotten stuck!')

        # Compute time series for each atomic predicate
        simulation_data = {}
        for name, predicate in self.predicates.items():
            series = [
                (time, predicate(lat, lon, psi, cte, he))
                for time, lat, lon, psi, cte, he in zip(times, lats, lons, psis, ctes, hes)
            ]
            simulation_data[name] = series
        return simulation_data


class XPlaneFalsifier(verifai.falsifier.mtl_falsifier):
    def __init__(self, monitor, sampler_type=None, sampler=None, sample_space=None,
                 falsifier_params={}, server_options={}):
        super().__init__(monitor, sampler_type=sampler_type, sampler=sampler,
                         sample_space=sample_space,
                         falsifier_params=falsifier_params, server_options=server_options)
        self.verbosity = falsifier_params.get('verbosity', 0)
        video = falsifier_params.get('video')
        if video is not None:
            import utils.images       # will throw exception if required libraries not available
            self.video_threshold = video['threshold']
            self.video_framerate = video['framerate']
        else:
            self.video_threshold = None

    def init_server(self, server_options):
        samplingConfig = DotMap(sampler=self.sampler,
                                sampler_type=self.sampler_type,
                                sample_space=self.sample_space)
        self.server = XPlaneServer(samplingConfig, self.monitor, server_options)

    def populate_error_table(self, sample, rho, error=True):
        super().populate_error_table(sample, rho, error)
        if self.video_threshold is not None and rho <= self.video_threshold:
            if error:
                index = len(self.error_table.table) - 1
                name = f'error-{index}'
            else:
                index = len(self.safe_table.table) - 1
                name = f'safe-{index}'
            import utils.images
            utils.images.write_video(self.server.images, filename=name+'.avi',
                                     fps=self.video_framerate)


def run_test(configuration, runway, verbosity=0):
    # Load Scenic scenario
    print('Loading scenario...')
    sampler = verifai.ScenicSampler.fromScenario(configuration['scenario'])

    # Define predicates and specifications
    def nearcenterline(lat, lon, psi, cte, he):
        """MTL predicate 'cte < 1.5', i.e. within 1.5 m of centerline."""
        return 1.5 - abs(cte)
    predicates = { 'nearcenterline': nearcenterline }
    specification = [configuration.get('specification', 'G nearcenterline')]

    # Set up controller
    framerate = configuration['framerate']
    controller = simple_controller.control if configuration['controller'] else None

    # Get options for video recording
    video = configuration.get('video')
    if video is None or not video['record']:
        video = None
    else:
        video['framerate'] = framerate

    # Create falsifier (and underlying server)
    serverOptions = DotMap(port=8080, controller=controller, framerate=framerate,
                           predicates=predicates, verbosity=verbosity,
                           runway=runway, video=video)
    falsifierOptions = DotMap(
        n_iters=configuration['runs'],
        verbosity=verbosity, video=video,
        save_error_table=True, save_safe_table=True,
        error_table_path=configuration['error_table'],
        safe_table_path=configuration['safe_table'],
    )
    falsifier = XPlaneFalsifier(specification, sampler=sampler,
                                falsifier_params=falsifierOptions,
                                server_options=serverOptions)

    # Perform falsification
    try:
        falsifier.run_falsifier()
    except socket.timeout as e:
        raise RuntimeError('lost connection to X-Plane') from e
    finally:
        # Print error and safe tables
        if verbosity >= 1:
            print('Error Table (also saved to {}):'.format(falsifier.error_table_path))
            print(falsifier.error_table.table)
            print('Safe Table (also saved to {}):'.format(falsifier.safe_table_path))
            print(falsifier.safe_table.table)
        # Write out final cross-entropy distributions
        es = sampler.scenario.externalSampler
        if es is not None:
            es = es.sampler.domainSampler
            if isinstance(es, verifai.samplers.cross_entropy.CrossEntropySampler):
                if verbosity >= 1:
                    print('Cross-entropy distributions:')
                with open(configuration['cross_entropy'], 'w') as outfile:
                    for s in es.split_sampler.samplers:
                        if verbosity >= 1:
                            print(s.dist)
                        outfile.write(str(s.dist)+'\n')


def load_yaml(filename):
    with open(filename, 'r') as stream:
        options = yaml.safe_load(stream)
    return options


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', help='experiment configuration file', default='config.yaml')
    parser.add_argument('-r', '--runway', help='runway configuration file', default='runway.yaml')
    parser.add_argument('-v', '--verbosity', type=int, default=0)
    args = parser.parse_args()

    # Parse runway configuration
    runway = load_yaml(args.runway)
    rads = runway['radians']
    runway_heading = runway['heading']
    if not rads:
        runway_heading = math.radians(runway_heading)
    runway_data = DotMap(
        heading=runway_heading, elevations=runway['elevations'],
        origin_x=runway['origin_X'], origin_z=runway['origin_Z'],
        start_lat=runway['start_lat'], start_lon=runway['start_lon'],
        end_lat=runway['end_lat'], end_lon=runway['end_lon']
    )

    # Parse experiment configuration
    configuration = load_yaml(args.config)

    run_test(configuration, runway_data, verbosity=args.verbosity)



"""Cross-entropy samplers"""

import numpy as np
from verifai.samplers.domain_sampler import BoxSampler, DiscreteBoxSampler, \
    DomainSampler, SplitSampler
from verifai.samplers.random_sampler import RandomSampler
import networkx as nx

class CrossEntropySampler(DomainSampler):
    def __init__(self, domain, ce_params):
        super().__init__(domain)
        self.alpha = ce_params.alpha
        self.thres = ce_params.thres
        self.cont_buckets = ce_params.cont.buckets
        self.cont_dist = ce_params.cont.dist
        self.disc_dist = ce_params.disc.dist
        self.cont_ce = lambda domain: ContinuousCrossEntropySampler(domain=domain,
                                                     buckets=self.cont_buckets,
                                                     dist=self.cont_dist,
                                                     alpha=self.alpha,
                                                     thres=self.thres)
        self.disc_ce = lambda domain: DiscreteCrossEntropySampler(domain=domain,
                                                   dist=self.disc_dist,
                                                   alpha=self.alpha,
                                                   thres=self.thres)
        partition = (
            (lambda d: d.standardizedDimension > 0, self.cont_ce),
            (lambda d: d.standardizedIntervals, self.disc_ce)
        )
        self.split_sampler = SplitSampler.fromPartition(domain,
                                                        partition,
                                                        RandomSampler)
        self.cont_sampler, self.disc_sampler = None, None
        self.rand_sampler = None
        for subsampler in self.split_sampler.samplers:
            if isinstance(subsampler, ContinuousCrossEntropySampler):
                assert self.cont_sampler is None
                self.cont_sampler = subsampler
            elif isinstance(subsampler, DiscreteCrossEntropySampler):
                assert self.disc_sampler is None
                self.disc_sampler = subsampler
            else:
                assert isinstance(subsampler, RandomSampler)
                assert self.rand_sampler is None
                self.rand_sampler = subsampler

    def getSample(self):
        return self.split_sampler.getSample()

    def update(self, sample, info, rho):
        self.split_sampler.update(sample, info, rho)

class ContinuousCrossEntropySampler(BoxSampler):
    def __init__(self, domain, alpha, thres,
                 buckets=10, dist=None):
        super().__init__(domain)
        if isinstance(buckets, int):
            buckets = np.ones(self.dimension) * buckets
        elif len(buckets) > 1:
            assert len(buckets) == self.dimension
        else:
            buckets = np.ones(self.dimension) * buckets[0]
        if dist is not None:
            assert (len(dist) == len(buckets))
        if dist is None:
            dist = [np.ones(int(b))/b for b in buckets]
        self.buckets = buckets
        self.dist = dist
        self.alpha = alpha
        self.thres = thres
        self.current_sample = None

    def getVector(self):
        bucket_samples = np.array([np.random.choice(int(b), p=self.dist[i])
                                   for i, b in enumerate(self.buckets)])
        self.current_sample = bucket_samples
        ret = tuple(np.random.uniform(bs, bs+1.)/b for b, bs
              in zip(self.buckets, bucket_samples))
        return ret, bucket_samples

    def updateVector(self, vector, info, rho):
        if rho is None or rho >= self.thres:
            return
        for row, b in zip(self.dist, info):
            row *= self.alpha
            row[b] += 1 - self.alpha

class DiscreteCrossEntropySampler(DiscreteBoxSampler):
    def __init__(self, domain, alpha, thres, dist=None):
        super().__init__(domain)
        if dist is not None:
            assert (len(dist) == len(domain.standardizedIntervals))
        if dist is None:
            dist = [np.ones(right-left+1)/(right-left+1)
                    for left, right in domain.standardizedIntervals]
        self.dist = dist
        self.alpha = alpha
        self.thres = thres
        self.current_sample = None

    def getVector(self):
        self.current_sample=\
            tuple(left + np.random.choice(right-left+1, p=self.dist[i])
                     for i, (left, right) in enumerate(self.domain.standardizedIntervals))
        return self.current_sample, None

    def updateVector(self, vector, info, rho):
        assert rho is not None
        if rho >= self.thres:
            return
        for row, (left, right), b in zip(self.dist, self.domain.standardizedIntervals, vector):
            row *= self.alpha
            row[b-left] += 1 - self.alpha

class MultiContinuousCrossEntropySampler(ContinuousCrossEntropySampler):
    
    def __init__(self, domain, alpha, thres, priority_graph=None,
                 buckets=10, dist=None, epsilon=0.5):
        if priority_graph is not None:
            self.set_graph(priority_graph)
        self.epsilon = epsilon
        self.still_sampling = False
        super().__init__(domain, alpha, thres, buckets=10, dist=dist)
        self.counts = np.array([np.zeros(int(b)) for b in self.buckets])

    def getVector(self):
        if not self.still_sampling:
            self.sample_randomly = np.random.uniform() < self.epsilon
        if self.sample_randomly:
            bucket_samples = np.array([np.random.choice(int(b))
                                    for i, b in enumerate(self.buckets)])
        else:
            bucket_samples = np.array([np.random.choice(int(b), p=self.dist[i])
                                    for i, b in enumerate(self.buckets)])
        self.current_sample = bucket_samples
        ret = tuple(np.random.uniform(bs, bs+1.)/b for b, bs
              in zip(self.buckets, bucket_samples))
        return ret, bucket_samples

    def set_graph(self, graph):
        self.priority_graph = graph
        if graph is not None:
            self.thres = [self.thres] * graph.number_of_nodes()
            self.num_properties = graph.number_of_nodes()

    def updateVector(self, vector, info, rho):
        if isinstance(rho, int):
            self.still_sampling = True
            return
        if rho is None or any([r is None for r in rho]):
            self.still_sampling = False
            return
        to_update = [True] * self.num_properties
        for node in nx.dfs_preorder_nodes(self.priority_graph):
            if not to_update[node]:
                continue
            b = rho[node] <= self.thres[node]
            if not b:
                to_update[node] = False
                for subnode in nx.descendants(self.priority_graph, node):
                    to_update[subnode] = False
        num_updates = sum(to_update)
        for crow, drow, b in zip(self.counts, self.dist, info):
            crow[b] += 1
            for _ in range(num_updates):
                row *= self.alpha
                row[b] += 1 - self.alpha



import numpy as np
import pygame
import carla
from .carla_world import *


class carla_task():
    def __init__(self,
                 n_sim_steps=500,
                 display_dim=(1280,720),
                 carla_host='127.0.0.1',
                 carla_port=2000,
                 carla_timeout=4.0,
                 world_map='Town03',
                 cam_transform=0):
        self.n_sim_steps = n_sim_steps
        self.display_dim = display_dim
        self.client = carla.Client(carla_host, carla_port)
        self.client.set_timeout(carla_timeout)
        self.clock = pygame.time.Clock()
        self.world_map = world_map
        self.timestep = 0
        self.cam_transform = cam_transform
        print("[carla_task] Finished initializing carla task.")

    def step_world(self):
        # print("[carla_task] Stepping world.")
        self.world.world.wait_for_tick()
        # print("[carla_task] Before tick")
        self.world.tick(self.clock)
        # print("[carla_task] After tick")
        self.world.render(self.display)
        # print("[carla_task] After render")
        self.client.apply_batch(self.world.get_control_cmds())
        # print("[carla_task] Batch applied.")
        pygame.display.flip()


    def run_task(self, sample):
        try:
            pygame.init()
            pygame.font.init()
            self.hud = HUD(*self.display_dim)
            self.display = pygame.display.set_mode(
                self.display_dim,
                pygame.HWSURFACE | pygame.DOUBLEBUF
            )
            # print("[carla_task] Setting up world.")
            if self.client.get_world().get_map().name == self.world_map:
                self.world = World(self.client.get_world(), self.hud, 
                        cam_transform=self.cam_transform)
            else:
                self.world = World(self.client.load_world(self.world_map), 
                        self.hud, cam_transform=self.cam_transform)
            # print("[carla_task] World setup complete.")
            self.use_sample(sample)
            self.world.restart()
            self.timestep = 0
            while self.timestep < self.n_sim_steps:
                self.step_world()
                self.timestep += 1
            traj = self.trajectory_definition()
        finally:
            self.world.destroy()
            pygame.quit()
        return traj


    def use_sample(self, sample):
        raise NotImplementedError('Method should be implemented in subclass.')


    def trajectory_definition(self):
        raise NotImplementedError('Method should be implemented in subclass.')



class webots_task():
    def __init__(self, N_SIM_STEPS, supervisor):
        self.N_SIM_STEPS = N_SIM_STEPS
        self.supervisor = supervisor

    def use_sample(self, sample):
        raise NotImplementedError("Define in child class")

    def run_task(self, sample):
        raise NotImplementedError("Define in child class")

    def close(self):
        self.supervisor.simulationReset()



import numpy as np

class dynamics:
    def __init__(self, nx, nu, f, dt = 0.1):
        self.nx = nx
        self.nu = nu
        self.f = f
        self.dt = dt
    def next_state(self, x, u):
        return x + self.f(x, u) * self.dt

class car_dynamics(dynamics):
    def __init__(self, wheelbase, dt=0.1):
        self.wheelbase = wheelbase

        def f(x, u):
            return np.array([x[2] * np.sin(float(np.pi/2 - x[3])),
                       x[2] * np.cos(float(np.pi/2 - x[3])),
                       u[1],
                       -x[2] * np.tan(float(u[0])) / self.wheelbase])

        super().__init__(nx=4, nu=2, f=f, dt=dt)






"""Specialized server for using Scenic's dynamic simulator interfaces."""

import time

from dotmap import DotMap
import progressbar

try:
    import ray
except ModuleNotFoundError:
    ray = None   # ignore for now; we'll raise an error below if ray is actually needed

from verifai.server import Server, ServerTimings
from verifai.samplers.scenic_sampler import ScenicSampler
from verifai.monitor import multi_objective_monitor
from scenic.core.simulators import SimulationCreationError
from scenic.core.external_params import VerifaiSampler
from scenic.core.distributions import RejectionException

class ScenicServer(Server):
    """`Server` for use with dynamic Scenic scenarios.

    Supported server options:

        * ``maxSteps``: maximum number of time steps to run a simulation;
        * ``verbosity``: verbosity level (as in the Scenic :option:`--verbosity` option);
        * ``maxIterations``: maximum number of iterations for rejection sampling;
        * ``simulator``: Scenic :obj:`~scenic.core.simulators.Simulator` to use, or
          `None` (the default) to use one specified in the scenario.
    """
    def __init__(self, sampling_data, monitor, options={}):
        if sampling_data.sampler is None:
            raise RuntimeError('ScenicServer created without sampler')
        self.sampler = sampling_data.sampler
        if not isinstance(self.sampler, ScenicSampler):
            raise RuntimeError('only a ScenicSampler can be used with ScenicServer')
        self.sample_space = self.sampler.space
        extSampler = self.sampler.scenario.externalSampler
        if extSampler is None:
            self.rejectionFeedback = None
        else:
            self.rejectionFeedback = extSampler.rejectionFeedback
        self.monitor = monitor
        self.lastValue = None
        defaults = DotMap(maxSteps=None, verbosity=0, maxIterations=1, simulator=None)
        defaults.update(options)
        self.maxSteps = defaults.maxSteps
        self.verbosity = defaults.verbosity
        self.maxIterations = defaults.maxIterations
        if defaults.simulator is None:
            self.simulator = self.sampler.scenario.getSimulator()
        else:
            self.simulator = defaults.simulator

    def evaluate_sample(self, sample):
        scene = self.sampler.lastScene
        assert scene
        result = self._simulate(scene)
        if result is None:
            return self.rejectionFeedback
        value = (0 if self.monitor is None
                 else self.monitor.evaluate(result))
        return value

    def _simulate(self, scene):
        startTime = time.time()
        if self.verbosity >= 1:
            print('  Beginning simulation...')
        try:
            result = self.simulator.simulate(scene,
                maxSteps=self.maxSteps, verbosity=self.verbosity,
                maxIterations=self.maxIterations)
        except SimulationCreationError as e:
            if self.verbosity >= 1:
                print(f'  Failed to create simulation: {e}')
            return None
        if self.verbosity >= 1:
            totalTime = time.time() - startTime
            print(f'  Ran simulation in {totalTime:.4g} seconds.')
        return result

    def terminate(self):
        pass

class DummySampler(VerifaiSampler):

    def nextSample(self, feedback):
        return self.last_sample

if ray:
    @ray.remote
    class SampleSimulator():

        def __init__(self, scenic_path, worker_num, monitor, options={}):
            scenario_params = options.get('scenario_params', {})
            scenario_model = options.get('scenario_model', None)
            scenario_params.update({
                'port': 2000 + 2*worker_num
            })
            self.sampler = ScenicSampler.fromScenario(scenic_path, maxIterations=1,
                                                      params=scenario_params,
                                                      model=scenario_model)
            # reset self.sampler.scenario.externalSampler to dummy sampler
            # that reads argument
            self.worker_num = worker_num
            self.sampler.scenario.externalSampler = DummySampler(
                self.sampler.scenario.externalParams,
                self.sampler.scenario.params
            )
            self.simulator = self.sampler.scenario.getSimulator()
            self.monitor = monitor
            extSampler = self.sampler.scenario.externalSampler
            if extSampler is None:
                self.rejectionFeedback = None
            else:
                self.rejectionFeedback = extSampler.rejectionFeedback
            defaults = DotMap(maxSteps=None, verbosity=0, maxIterations=1)
            defaults.update(options)
            self.maxSteps = defaults.maxSteps
            self.verbosity = defaults.verbosity
            self.maxIterations = defaults.maxIterations

        def get_sample(self, sample):
            self.sampler.scenario.externalSampler.last_sample = sample
            self.full_sample = self.sampler.nextSample(sample)

        def simulate(self, sample):
            '''
            Need to generate scene from sample here.
            '''
            t0 = time.time()
            self.sampler.scenario.externalSampler.last_sample = sample
            scene = self.sampler.lastScene
            startTime = time.time()
            if self.verbosity >= 1:
                print('  Beginning simulation...')
            try:
                result = self.simulator.simulate(scene,
                    maxSteps=self.maxSteps, verbosity=self.verbosity,
                    maxIterations=self.maxIterations)
                result.worker_num = self.worker_num
            except SimulationCreationError as e:
                if self.verbosity >= 1:
                    print(f'  Failed to create simulation: {e}')
                self.lastValue = self.rejectionFeedback
                return self.worker_num, self.full_sample, self.lastValue
            except RuntimeError as e:
                print(f'Runtime error during simulation: {e}')
                print('Waiting 1 minute before continuing...')
                time.sleep(60)
                return self.worker_num, self.full_sample, self.lastValue
            if self.verbosity >= 1:
                totalTime = time.time() - startTime
                print(f'  Ran simulation in {totalTime:.4g} seconds.')
            if result is None:
                self.lastValue = self.rejectionFeedback
            else:
                self.lastValue = (0 if self.monitor is None
                                  else self.monitor.evaluate(result))
            return self.worker_num, self.full_sample, self.lastValue

class ParallelScenicServer(ScenicServer):

    def __init__(self, total_workers, n_iters, sampling_data, scenic_path, monitor,
                 options={}, max_time=None, sampler=None):
        if not ray:
            raise RuntimeError('ParallelScenicServer requires ray to be installed')

        if not ray.is_initialized():
            ray.init(ignore_reinit_error=True)
        self.total_workers = total_workers
        self.n_iters = n_iters
        self.max_time = max_time
        sampling_data.sampler = sampler
        super().__init__(sampling_data, monitor, options)
        self.sample_simulators = [
            SampleSimulator.remote(scenic_path, i, monitor, options)
            for i in range(self.total_workers)
        ]

    def _generate_next_sample(self, worker_num):
        i = 0
        ext = self.sampler.scenario.externalSampler
        while i < 2000:
            ext.cachedSample, info = ext.getSample()
            sample = ext.cachedSample
            sim = self.sample_simulators[worker_num]
            try:
                ray.get(sim.get_sample.remote(sample))
                return sample, info
            except SimulationCreationError as e:
                if self.verbosity >= 1:
                    print(f'  Failed to create simulation: {e}')
                return None, None
            except RejectionException as e:
                i += 1
                continue
        return None, None

    def run_server(self):
        results = []
        futures = []
        samples = []
        infos = []
        if self.n_iters is not None:
            bar = progressbar.ProgressBar(max_value=self.n_iters)
        else:
            widgets = ['Scenes generated: ', progressbar.Counter('%(value)d'),
               ' (', progressbar.Timer(), ')']
            bar = progressbar.ProgressBar(widgets=widgets)
        for i in range(self.total_workers):
            next_sample, info = self._generate_next_sample(i)
            samples.append(next_sample)
            infos.append(info)
            sim = self.sample_simulators[i]
            futures.append(sim.simulate.remote(next_sample))
        while True:
            done, _ = ray.wait(futures)
            result = ray.get(done[0])
            index, sample, rho = result
            self.lastValue = rho
            results.append((sample, rho))
            info = infos[index]
            self.sampler.scenario.externalSampler.update(sample, info, rho)
            bar.update(len(results))
            if len(results) == 1:
                t0 = time.time()
            elapsed = time.time() - t0
            if self.n_iters is not None and len(results) >= self.n_iters:
                break
            if self.max_time is not None and elapsed >= self.max_time:
                break
            next_sample, info = self._generate_next_sample(index)
            elapsed = time.time() - t0
            sim = self.sample_simulators[index]
            samples[index] = next_sample
            infos[index] = info
            futures[index] = sim.simulate.remote(next_sample)

        return results



import os

import numpy as np

try:
    import mss
    import cv2
except ImportError as e:
    raise RuntimeError('recording images requires the mss and cv2 packages') from e

sct = mss.mss()

def grab_image(monitor, resizeTo=None):
    img = cv2.cvtColor(np.array(sct.grab(monitor)), cv2.COLOR_BGRA2BGR)
    if resizeTo is not None:
        return cv2.resize(img, resizeTo)
    return img

def write_video(images, filename='out.avi', fps=10.0):
    height, width, _ = images[0].shape
    fourcc = cv2.VideoWriter_fourcc(*'XVID')
    writer = cv2.VideoWriter(filename, fourcc, fps, (width, height))
    if not writer.isOpened():
        raise RuntimeError('failed to initialize VideoWriter')
    try:
        for image in images:
            writer.write(image)
    finally:
        writer.release()




"""Samplers generating points in a Domain, possibly subject to specifications.
"""

### Exceptions pertaining to sampling

class SamplingError(Exception):
    """Exception raised if a Sampler is unable to produce a sample"""
    pass

### Exceptions pertaining to sampling termination

class TerminationException(Exception):
    """Exception raised if sampling has terminated,
    e.g. grid sampling and bayesian optimization"""
    pass

### Samplers defined over fixed Domains

## Abstract samplers

class DomainSampler:
    """Abstract sampler class"""

    def __init__(self, domain):
        self.domain = domain
        self.last_sample = None

    def getSample(self):
        """Generate the next sample, given the current distribution.

        Must return a pair consisting of the sample and arbitrary
        sampler-specific info, which will be passed to the `update`
        method after the sample is evaluated.
        """
        raise NotImplementedError('tried to use abstract Sampler')
    
    def update(self, sample, info, rho):
        """
        Use the provided sample and rho value to update the state of the sampler.

        Arguments:
            sample: the exact value generated by the sampler.
            info: any additional info needed to complete the update,
            such as the buckets that were sampled from for continuous samplers.
            rho: the result of the sample, returned by the falsifier.
        """
        pass

    def nextSample(self, feedback=None):
        """Generate the next sample, given feedback from the last sample.

        This exists only for backwards-compatibility. It has been replaced by
        the getSample and update APIs.
        """
        if self.last_sample is not None:
            self.update(self.last_sample, self.last_info, feedback)
        self.last_sample, self.last_info = self.getSample()
        return self.last_sample

    def __iter__(self):
        try:
            feedback = None
            while True:
                feedback = yield self.nextSample(feedback)
        except TerminationException:
            return

class ConstrainedSampler(DomainSampler):
    """Abstract DomainSampler constrained by a specification."""

    def __init__(self, domain, spec):
        super().__init__(domain)
        self.specification = spec

class SplitSampler(DomainSampler):
    """Sampler using different strategies for different types of domains"""

    def __init__(self, domain, samplers):
        super().__init__(domain)
        self.samplers = tuple(samplers)

    def getSample(self):
        samples, infos = [], []
        for sampler in self.samplers:
            sample, info = sampler.getSample()
            samples.append(sample)
            infos.append(info)
        return self.domain.rejoinPoints(*samples), (samples, infos)

    def update(self, sample, info, rho):
        samples, infos = info
        for sampler, subsample, i in zip(self.samplers, samples, infos):
            sampler.update(subsample, i, rho)

    @classmethod
    def fromPartition(cls, domain, partition, defaultSampler=None):
        """Make a SplitSampler by partitioning a domain by predicates.

        The given partition should consist of a sequence of pairs of predicates
        and functions which return a sampler given a domain. The part of the
        domain satisfying the first predicate is given to the first function to
        make a sampler for it. The remaining part of the space is split along
        the second predicate, and so forth. If the partition is non-exhaustive
        (part of the domain does not satisfy any of the predicates), then an
        error is raised unless a default sampler is specified, in which case it
        is called with the remaining part of the space to make a sampler.

        This sampler this method returns has a samplersForPredicates attribute
        which indicates for the ith predicate which subsampler (or None)
        corresponds to its part of the domain. If a default sampler is used,
        the tuple has an extra last element for it.
        """
        samplers = []
        remaining = domain
        samplersForPredicates = [None for i in partition]
        for i, (predicate, makeSampler) in enumerate(partition):
            component, remaining = remaining.partition(predicate)
            if component:
                sampler = makeSampler(component)
                samplers.append(sampler)
                samplersForPredicates[i] = sampler
            if not remaining:
                break
        if remaining:
            if defaultSampler is None:
                raise RuntimeError('tried to make SplitSampler with'
                                   ' non-exhaustive partition')
            else:
                sampler = defaultSampler(remaining)
                samplers.append(sampler)
                samplersForPredicates.append(sampler)
        assert len(samplers) > 0
        sampler = cls(domain, samplers)
        sampler.samplersForPredicates = tuple(samplersForPredicates)
        return sampler

    @classmethod
    def fromPredicate(cls, domain, predicate, leftSampler, rightSampler):
        """Make a SplitSampler by splitting a domain by a single predicate."""
        return cls.fromPartition(domain,
                                 ((predicate, leftSampler),),
                                 rightSampler)

class BoxSampler(DomainSampler):
    """Samplers defined only over unit hyperboxes"""
    def __init__(self, domain):
        self.dimension = domain.standardizedDimension
        if not self.dimension >= 0:
            raise RuntimeError(f'{self.__class__.__name__} supports only'
                               ' continuous standardizable Domains')
        super().__init__(domain)
    
    def getSample(self):
        sample, info = self.getVector()
        return self.domain.unstandardize(sample), info

    def getVector(self):
        raise NotImplementedError('tried to use abstract BoxSampler')

    def update(self, sample, info, rho):
        self.updateVector(self.domain.standardize(sample), info, rho)

    def updateVector(self, vector, info, rho):
        pass

    def set_graph(self, graph):
        self.priority_graph = graph
        if graph is not None:
            self.thres = [self.thres] * graph.number_of_nodes()
            self.num_properties = graph.number_of_nodes()
            self.is_multi = True

class DiscreteBoxSampler(DomainSampler):
    """Samplers defined only over discrete hyperboxes"""
    def __init__(self, domain):
        self.intervals = domain.standardizedIntervals
        if not self.intervals:
            raise RuntimeError(f'{self.__class__.__name__} supports only'
                               ' discrete standardizable Domains')
        super().__init__(domain)

    def getSample(self):
        sample, info = self.getVector()
        return self.domain.unstandardize(sample), info

    def getVector(self):
        raise NotImplementedError('tried to use abstract DiscreteBoxSampler')

    def update(self, sample, info, rho):
        self.updateVector(self.domain.standardize(sample), info, rho)

    def updateVector(self, vector, info, rho):
        pass

class IteratorSampler(DomainSampler):
    """Samplers defined using a generator function."""

    def __init__(self, domain, repeat=False):
        super().__init__(domain)
        self.repeat = repeat
        self.iterator = iter(self)
        self.lastFeedback = None
        self.waitingForFeedback = False

    def getSample(self):
        if self.waitingForFeedback:
            raise RuntimeError('IteratorSampler does not support parallel sampling')
        self.waitingForFeedback = True
        try:
            return self.iterator.send(self.lastFeedback), None
        except StopIteration:
            if self.repeat:
                self.iterator = iter(self)
                return self.iterator.send(self.lastFeedback), None
            else:
                raise TerminationException from None

    def update(self, sample, info, rho):
        self.lastFeedback = rho
        self.waitingForFeedback = False

    def __iter__(self):
        raise NotImplementedError('tried to iterate abstract IteratorSampler')



"""Example aircraft controller.

This controller simply applies a constant throttle, and uses a proportional
controller for the rudder based on the ground truth cross-track and heading
errors.
"""

throttle = 0.4  # constant throttle of plane
cte_gain = 0.1  # gain for cross-track error
he_gain = 0.05  # gain for heading error

def control(server, lat, lon, psi, cte, heading_err):
    rudder = (cte_gain * cte) + (he_gain * heading_err)
    server.sendCTRL([0.0, 0.0, rudder, throttle])




"""Features, feature spaces, and domains.

.. testsetup::
    from verifai.features import *
"""

import math
import random
import itertools
import functools
from collections import OrderedDict, namedtuple

import numpy as np

from verifai.utils.utils import RejectionException, cached_property

### Domains

class Domain:
    """Abstract class of domains"""

    def distance(self, pointA, pointB):
        raise NotImplementedError('domain distance metric not implemented')

    def uniformPoint(self):
        """Sample a uniformly random point in this Domain"""
        raise NotImplementedError('domain uniform sampling not implemented')

    def flatten(self, point):
        """Flatten a point in this Domain to a tuple of coordinates.

        Useful for analyses that do not understand the internal structure of
        Domains. This representation of a point is also hashable, and so can be
        put into sets and dicts.
        """
        coords = []
        self.flattenOnto(point, coords)
        return tuple(coords)

    def flattenOnto(self, point, targetList):
        """Flatten a point onto the end of the given list."""
        raise NotImplementedError('domain flattening not implemented')

    @cached_property
    def flattenedDimension(self):
        """Dimension of the vector returned by flatten."""
        return NotImplementedError('domain flattening not implemented')

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        """Meaning of a coordinate of a flattened point in this Domain.

        If pointName is the name of a variable storing a point in the Domain,
        then this function returns an expression extracting from that variable
        the value which would be stored in the given coordinate index when the
        point is flattened. For example:

        >>> struct = Struct({'a': Real(), 'b': Real()})
        >>> point = struct.makePoint(a=4, b=3)
        >>> struct.flatten(point)
        (4.0, 3.0)
        >>> struct.meaningOfFlatCoordinate(1)
        'point.b'
        >>> eval(struct.meaningOfFlatCoordinate(1))
        3
        """
        return NotImplementedError('domain flattening not implemented')

    def pandasIndexForFlatCoordinate(self, index):
        """Like meaningOfFlatCoordinate, but giving a MultiIndex for pandas."""
        return NotImplementedError('domain flattening not implemented')

    def coordinateIsNumerical(self, index):
        """Whether the value of a coordinate is intrinsically numerical.

        In particular, whether it makes sense to use the Euclidean distance
        between different values of the coordinate. This would not be the case
        for Domains whose points are strings, for example, even if those are
        converted to numbers for the purpose of flattening.
        """
        return NotImplementedError('domain flattening not implemented')

    def numericizeCoordinate(self, coord):
        """Make a coordinate numeric. For internal use."""
        return coord

    def denumericizeCoordinate(self, coord):
        """Reconstitute a coordinate's original value. For internal use."""
        return coord

    def unflatten(self, coords):
        """Unflatten a tuple of coordinates to a point in this Domain."""
        iterator = iter(coords)
        try:
            return self.unflattenIterator(iterator)
        except StopIteration as e:
            raise RuntimeError('malformed flattened point') from e

    def unflattenIterator(self, coords):
        """Unflatten an iterator of coordinates to a point in this Domain."""
        raise NotImplementedError('domain unflattening not implemented')

    def standardize(self, point):
        """Map the point into a hyperbox, preserving measure.

        If the Domain is continuous, this should map into a unit hyperbox.
        If it is discrete, it should map into a discrete hyperbox.
        Which (if either) of these is the case can be determined by calling
        standardizedDimension and standardizedIntervals: for primitive Domains,
        at least one will return the 'not supported' value."""
        coords = []
        self.standardizeOnto(point, coords)
        return tuple(coords)

    @cached_property
    def isStandardizable(self):
        """Whether points in this Domain can be standardized as above."""
        return self.standardizedDimension >= 0 or self.standardizedIntervals

    def standardizeOnto(self, point, targetList):
        """Standardize a point onto the end of the given list."""
        raise RuntimeError(
            f'Domain {self.__class__.__name__} does not support standardize')

    @cached_property
    def standardizedDimension(self):
        """Dimension of the vector returned by standardize.

        Returns -1 if continuous standardization is not supported."""
        return -1

    @cached_property
    def standardizedIntervals(self):
        """Intervals of the DiscreteBox mapped into by standardize.

        Returns () if discrete standardization is not supported."""
        return ()

    def unstandardize(self, coords):
        """Unstandardize a tuple of coordinates to a point in this Domain."""
        iterator = iter(coords)
        try:
            return self.unstandardizeIterator(iterator)
        except StopIteration as e:
            raise RuntimeError('malformed standardized point') from e

    def unstandardizeIterator(self, coords):
        """Unstandardize an iterator of coords to a point in this Domain."""
        raise RuntimeError(
            f'Domain {self.__class__.__name__} does not support standardize')

    @cached_property
    def requiresRejection(self):
        """Whether sampling from this Domain requires rejection sampling."""
        return False

    def partition(self, predicate):
        """Split this Domain into parts satisfying/falsifying the predicate."""
        if predicate(self):
            return self, None
        else:
            return None, self

    def rejoinPoints(self, *components):
        """Join points from the partitioned components of a Domain."""
        assert sum(int(component is not None) for component in components) == 1
        for component in components:
            if component is not None:
                return component

    @cached_property
    def pointsAreScalars(self):
        """Whether points in this Domain are scalars."""
        return False

## Primitive domains

class Constant(Domain):
    """Domain consisting of a single value"""

    def __init__(self, value):
        try:
            hash(value)
        except TypeError as e:
            raise RuntimeError('value for Constant domain '
                               f'is not hashable: {value}') from e
        self.value = value

    def distance(self, pointA, pointB):
        return 0

    def uniformPoint(self):
        return self.value

    def flattenOnto(self, point, targetList):
        pass    # do nothing: no point in encoding the constant value

    @cached_property
    def flattenedDimension(self):
        return 0

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        raise RuntimeError('Constant domain has no coordinates')

    def pandasIndexForFlatCoordinate(self, index):
        raise RuntimeError('Constant domain has no coordinates')

    def coordinateIsNumerical(self, index):
        raise RuntimeError('Constant domain has no coordinates')

    def unflattenIterator(self, coords):
        return self.value

    def __eq__(self, other):
        return (type(other) is Constant and self.value == other.value)

    def __hash__(self):
        return hash(self.value)

    def __iter__(self):
        yield self.value

    def __repr__(self):
        return f'Constant({self.value})'

class Categorical(Domain):
    """Domain consisting of a finite set of values"""

    def __init__(self, *values):
        self.values = tuple(values)
        if len(self.values) == 0:
            raise RuntimeError('Categorical must contain at least one value')
        try:
            self.index = { val: ind for ind, val in enumerate(self.values) }
        except TypeError as e:
            raise RuntimeError('points in a Domain must be hashable') from e
        if len(self.index) != len(self.values):
            raise RuntimeError('duplicate values in Categorical domain')

    def distance(self, pointA, pointB):
        return 0 if pointA == pointB else 1

    def uniformPoint(self):
        return random.choice(self.values)

    def flattenOnto(self, point, targetList):
        targetList.append(self.numericizeCoordinate(point))

    @cached_property
    def flattenedDimension(self):
        return 1

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        assert index == 0
        return pointName

    def pandasIndexForFlatCoordinate(self, index):
        assert index == 0
        return ()

    def coordinateIsNumerical(self, index):
        assert index == 0
        return False

    def numericizeCoordinate(self, coord):
        return self.index[coord]

    def denumericizeCoordinate(self, coord):
        return self.values[coord]

    def unflattenIterator(self, coords):
        return self.denumericizeCoordinate(next(coords))

    def standardizeOnto(self, point, targetList):
        targetList.append(self.numericizeCoordinate(point))

    @cached_property
    def standardizedIntervals(self):
        return ((0, len(self.values)-1),)

    def unstandardizeIterator(self, coords):
        return self.denumericizeCoordinate(next(coords))

    @cached_property
    def pointsAreScalars(self):
        return True     # TODO hmm

    def __eq__(self, other):
        if type(other) is not Categorical:
            return False
        return set(self.values) == set(other.values)

    def __hash__(self):
        return hash(self.values)

    def __iter__(self):
        yield from self.values

    def __repr__(self):
        return f'Categorical({self.values})'

class Real(Domain):
    """Domain of real numbers"""

    def distance(self, pointA, pointB):
        return math.abs(pointA - pointB)

    def uniformPoint(self):
        raise RuntimeError('cannot sample uniformly from all real numbers')

    def flattenOnto(self, point, targetList):
        targetList.append(float(point))

    @cached_property
    def flattenedDimension(self):
        return 1

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        assert index == 0
        return pointName

    def pandasIndexForFlatCoordinate(self, index):
        assert index == 0
        return ()

    def coordinateIsNumerical(self, index):
        assert index == 0
        return True

    def unflattenIterator(self, coords):
        return next(coords)

    @cached_property
    def pointsAreScalars(self):
        return True

    def __eq__(self, other):
        return type(other) is Real

    def __hash__(self):
        return hash(Real)

    def __repr__(self):
        return 'Real()'

class Integer(Domain):
    """Domain of integers"""

    def distance(self, pointA, pointB):
        return math.abs(pointA - pointB)

    def uniformPoint(self):
        raise RuntimeError('cannot sample uniformly from all integers')

    def flattenOnto(self, point, targetList):
        targetList.append(int(point))

    @cached_property
    def flattenedDimension(self):
        return 1

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        assert index == 0
        return pointName

    def pandasIndexForFlatCoordinate(self, index):
        assert index == 0
        return ()

    def coordinateIsNumerical(self, index):
        assert index == 0
        return True

    def unflattenIterator(self, coords):
        return next(coords)

    @cached_property
    def pointsAreScalars(self):
        return True

    def __eq__(self, other):
        return type(other) is Integer

    def __hash__(self):
        return hash(Integer)

    def __repr__(self):
        return 'Integer()'

class AbstractBox(Domain):
    """A hyper-box, i.e. a product of intervals"""

    def __init__(self, *intervals):
        """Create an AbstractBox from a sequence of intervals"""
        ints = tuple(intervals)
        dim = len(ints)
        if dim == 0:
            raise RuntimeError('attempted to create zero-dimensional Box')
        newInts = []
        lefts = []
        lengths = []
        for interval in ints:
            endpoints = tuple(interval)
            if len(endpoints) != 2:
                raise RuntimeError('interval for Box must have two endpoints')
            left, right = endpoints
            if left > right:
                left, right = right, left
            newInts.append((left, right))
            lefts.append(left)
            lengths.append(right - left)
        self.intervals = tuple(newInts)
        self.lefts = tuple(lefts)
        self.lengths = tuple(lengths)

    def distance(self, pointA, pointB):
        pa, pb = np.array(pointA), np.array(pointB)
        return np.linalg.norm(pa - pb)

    def flattenOnto(self, point, targetList):
        targetList.extend(point)

    @cached_property
    def flattenedDimension(self):
        return self.dimension

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        assert 0 <= index < self.dimension
        return f'{pointName}[{index}]'

    def pandasIndexForFlatCoordinate(self, index):
        assert 0 <= index < self.dimension
        return (index,)

    def coordinateIsNumerical(self, index):
        assert 0 <= index < self.dimension
        return True

    def unflattenIterator(self, coords):
        return tuple(itertools.islice(coords, self.dimension))

    @cached_property
    def pointsAreScalars(self):
        return self.dimension == 1

    @cached_property
    def dimension(self):
        return len(self.intervals)

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return self.intervals == other.intervals

    def __hash__(self):
        return hash(self.intervals)

class Box(AbstractBox):
    """A hyper-box over the reals.

    Points in a Box are tuples of floats.
    """

    def uniformPoint(self):
        return tuple(random.uniform(lo, hi) for lo, hi in self.intervals)

    def standardizeOnto(self, point, targetList):
        for coord, left, length in zip(point, self.lefts, self.lengths):
            targetList.append((coord - left) / length)

    @cached_property
    def standardizedDimension(self):
        return len(self.intervals)

    def unstandardizeIterator(self, coords):
        point = tuple(
            left + (coord * length)
            for left, length, coord in zip(self.lefts, self.lengths, coords)
        )
        assert len(point) == self.dimension
        return point

    def __repr__(self):
        args = ', '.join(repr(interval) for interval in self.intervals)
        return f'Box({args})'

class DiscreteBox(AbstractBox):
    """A hyper-box over the integers.

    Points in a DiscreteBox are tuples of ints.
    """

    def __init__(self, *intervals):
        super().__init__(*intervals)
        self.intervals = tuple((math.ceil(left), math.floor(right))
                               for left, right in self.intervals)
        for left, right in self.intervals:
            if left > right:
                raise RuntimeError('DiscreteBox interval contains no integers')

    def uniformPoint(self):
        return tuple(random.randint(lo, hi) for lo, hi in self.intervals)

    def standardizeOnto(self, point, targetList):
        targetList.extend(point)

    @cached_property
    def standardizedIntervals(self):
        return self.intervals

    def unstandardizeIterator(self, coords):
        return tuple(itertools.islice(coords, self.dimension))

    def __iter__(self):
        ranges = (range(left, right+1) for left, right in self.intervals)
        yield from itertools.product(*ranges)

    def __repr__(self):
        args = ', '.join(repr(interval) for interval in self.intervals)
        return f'DiscreteBox({args})'

## Domains constructed from other domains

class Array(Domain):
    """A multidimensional array of elements in a common domain.

    For example, Array(Box((-1, 1)), (10, 5)) represents a 10x5 grid
    of real numbers, each in the interval [-1, 1].

    Points in an Array are nested tuples of elements. For example, a point in
    the Array above would be a tuple of 10 elements, each of which is a tuple
    of 5 elements, each of which is a point in the underlying Box domain.
    """

    def __init__(self, domain, shape):
        self.domain = domain
        self.shape = tuple(map(int, shape))
        if len(self.shape) == 0:
            raise RuntimeError('Array shape must be nonempty')
        elts = 1
        for n in self.shape:
            if n < 0:
                raise RuntimeError('Array dimensions must be nonnegative')
            elif elts == 0:
                raise RuntimeError('only the last dimension of an Array can be zero')
            elts *= n
        self.numElements = elts

    def pointWithElements(self, it):
        """Build a point in this domain from an iterable of elements.

        This is similar to numpy.reshape, building a multidimensional array
        from a flat list of elements. For example:

        >>> elts = [1, 2, 3, 4, 5, 6]
        >>> array = Array(Real(), (2, 3))
        >>> array.pointWithElements(elts)
        ((1, 2, 3), (4, 5, 6))
        >>> array = Array(Real(), (3, 2))
        >>> array.pointWithElements(elts)
        ((1, 2), (3, 4), (5, 6))
        """
        it = iter(it)
        shape = self.shape
        dim = len(shape)
        def makeLevel(i):
            if i == dim:    # bottom level is a point in the underlying domain
                return next(it)
            else:   # upper levels are tuples of sublevels
                j = i + 1
                return tuple(makeLevel(j) for k in range(shape[i]))
        return makeLevel(0)

    def elementsOfPoint(self, point):
        """Return an iterator over the elements of a point in this domain.

        This is similar to numpy.flatten, turning a multidimensional array into
        a flat list of elements. For example:

        >>> array = Array(Real(), (3, 2))
        >>> list(array.elementsOfPoint(((1, 2), (3, 4), (5, 6))))
        [1, 2, 3, 4, 5, 6]
        """
        dim = len(self.shape)
        def iterateLevel(i, level):
            if i == dim:    # at bottom, yield the next element
                yield level
            else:   # recursively yield elements of each sublevel
                j = i + 1
                for sublevel in level:
                    yield from iterateLevel(j, sublevel)
        yield from iterateLevel(0, point)

    def uniformPoint(self):
        return self.pointWithElements(iter(self.domain.uniformPoint, None))

    def flattenOnto(self, point, targetList):
        for element in self.elementsOfPoint(point):
            self.domain.flattenOnto(element, targetList)

    @cached_property
    def flattenedDimension(self):
        return self.numElements * self.domain.flattenedDimension

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        assert 0 <= index < self.flattenedDimension
        ourIndex = index // self.domain.flattenedDimension
        subIndex = index % self.domain.flattenedDimension
        indices = []
        for levelSize in reversed(self.shape):
            levelIndex = ourIndex % levelSize
            indices.insert(0, f'[{levelIndex}]')
            ourIndex = ourIndex // levelSize
        meaning = pointName + ''.join(indices)
        return self.domain.meaningOfFlatCoordinate(subIndex, pointName=meaning)

    def pandasIndexForFlatCoordinate(self, index):
        assert 0 <= index < self.flattenedDimension
        ourIndex = index // self.domain.flattenedDimension
        subIndex = index % self.domain.flattenedDimension
        indices = []
        for levelSize in reversed(self.shape):
            levelIndex = ourIndex % levelSize
            indices.insert(0, levelIndex)
            ourIndex = ourIndex // levelSize
        subPandasIndex = self.domain.pandasIndexForFlatCoordinate(subIndex)
        return tuple(indices) + subPandasIndex

    def coordinateIsNumerical(self, index):
        assert 0 <= index < self.flattenedDimension
        subIndex = index % self.domain.flattenedDimension
        return self.domain.coordinateIsNumerical(subIndex)

    def unflattenIterator(self, coords):
        it = iter(lambda: self.domain.unflattenIterator(coords), None)
        return self.pointWithElements(it)

    def standardizeOnto(self, point, targetList):
        for element in self.elementsOfPoint(point):
            self.domain.standardizeOnto(element, targetList)

    @cached_property
    def standardizedDimension(self):
        sd = self.domain.standardizedDimension
        return self.numElements * sd if sd >= 0 else -1

    @cached_property
    def standardizedIntervals(self):
        return self.numElements * self.domain.standardizedIntervals

    def unstandardizeIterator(self, coords):
        it = iter(lambda: self.domain.unstandardizeIterator(coords), None)
        return self.pointWithElements(it)

    def partition(self, predicate):
        if self.numElements == 0:   # length-0 arrays count as leaves
            return super().partition(predicate)
        left, right = self.domain.partition(predicate)
        left = Array(left, self.shape) if left else None
        right = Array(right, self.shape) if right else None
        return left, right

    def rejoinPoints(self, *components):
        values = []
        assert any(component is not None for component in components)
        its = (itertools.repeat(None)
               if comp is None
               else self.elementsOfPoint(comp)
               for comp in components)
        for elements in zip(*its):
            values.append(self.domain.rejoinPoints(*elements))
        return self.pointWithElements(values)

    @cached_property
    def requiresRejection(self):
        return self.domain.requiresRejection

    @cached_property
    def pointsAreScalars(self):
        return self.shape == (1,)

    def __eq__(self, other):
        if not isinstance(other, Array):
            return NotImplemented
        return (self.domain == other.domain and self.shape == other.shape)

    def __hash__(self):
        return hash((self.domain, self.shape))

    def __iter__(self):
        for values in itertools.product(self.domain, repeat=self.numElements):
            yield self.pointWithElements(values)

    def __repr__(self):
        return f'Array({self.domain}, {self.shape})'

class ScalarArray(Array):
    """An array whose elements are integers or reals.

    This is a specialized implementation of Array optimized for large arrays of
    scalars like images.
    """

    def __init__(self, domain, shape):
        super().__init__(domain, shape)
        if not self.domain.pointsAreScalars:
            raise RuntimeError('ScalarArray element domain is not scalar')

    def flattenOnto(self, point, targetList):
        targetList.extend(np.ravel(point))

    @cached_property
    def flattenedDimension(self):
        return self.numElements

    def unflattenIterator(self, coords):
        values = list(itertools.islice(coords, self.numElements))
        return np.reshape(np.array(values), self.shape)

    def __repr__(self):
        return f'ScalarArray({self.domain}, {self.shape})'

class Struct(Domain):
    """A domain consisting of named sub-domains.

    The order of the sub-domains is arbitrary: two Structs are considered equal
    if they have the same named sub-domains, regardless of order. As the order
    is an implementation detail, accessing the values of sub-domains in points
    sampled from a Struct should be done by name:

    >>> struct = Struct({'a': Box((0, 1)), 'b': Box((2, 3))})
    >>> point = struct.uniformPoint()
    >>> point.b
    (2.20215292046797,)

    Within a given version of VerifAI, the sub-domain order is consistent, so
    that the order of columns in error tables is also consistent.
    """

    def __init__(self, domains):
        self.namedDomains = tuple(sorted(domains.items(), key=lambda i: i[0]))
        self.domainNamed = OrderedDict(self.namedDomains)
        self.domains = tuple(self.domainNamed.values())
        self.makePoint = namedtuple('StructPoint', self.domainNamed.keys())

    def uniformPoint(self):
        return self.makePoint(*(d.uniformPoint() for d in self.domains))

    def flattenOnto(self, point, targetList):
        for subPoint, domain in zip(point, self.domains):
            domain.flattenOnto(subPoint, targetList)

    @cached_property
    def flattenedDimension(self):
        return sum(domain.flattenedDimension for domain in self.domains)

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        assert 0 <= index < self.flattenedDimension
        for name, domain in self.namedDomains:
            if index < domain.flattenedDimension:
                subPoint = f'{pointName}.{name}'
                return domain.meaningOfFlatCoordinate(index,
                                                      pointName=subPoint)
            index -= domain.flattenedDimension

    def pandasIndexForFlatCoordinate(self, index):
        assert 0 <= index < self.flattenedDimension
        for name, domain in self.namedDomains:
            if index < domain.flattenedDimension:
                return (name,) + domain.pandasIndexForFlatCoordinate(index)
            index -= domain.flattenedDimension

    def coordinateIsNumerical(self, index):
        assert 0 <= index < self.flattenedDimension
        for name, domain in self.namedDomains:
            if index < domain.flattenedDimension:
                return domain.coordinateIsNumerical(index)
            index -= domain.flattenedDimension

    def unflattenIterator(self, coords):
        subPts = (domain.unflattenIterator(coords) for domain in self.domains)
        return self.makePoint(*subPts)

    def standardizeOnto(self, point, targetList):
        for subPoint, domain in zip(point, self.domains):
            domain.standardizeOnto(subPoint, targetList)

    @cached_property
    def standardizedDimension(self):
        total = 0
        for domain in self.domains:
            sd = domain.standardizedDimension
            if sd < 0:
                return -1
            else:
                total += sd
        return total

    @cached_property
    def standardizedIntervals(self):
        ints = []
        for domain in self.domains:
            subints = domain.standardizedIntervals
            if subints == ():
                return ()
            else:
                ints.extend(subints)
        return tuple(ints)

    def unstandardizeIterator(self, coords):
        subPts = (dom.unstandardizeIterator(coords) for dom in self.domains)
        return self.makePoint(*subPts)

    def partition(self, predicate):
        leftSubs, rightSubs = {}, {}
        for name, domain in self.namedDomains:
            left, right = domain.partition(predicate)
            if left:
                leftSubs[name] = left
            if right:
                rightSubs[name] = right
        left = Struct(leftSubs) if leftSubs else None
        right = Struct(rightSubs) if rightSubs else None
        return left, right

    def rejoinPoints(self, *components):
        subPoints = (
            domain.rejoinPoints(*(getattr(comp, name, None)
                                  for comp in components))
            for name, domain in self.namedDomains
        )
        return self.makePoint(*subPoints)

    @cached_property
    def requiresRejection(self):
        return any(domain.requiresRejection for domain in self.domains)

    def __iter__(self):
        for values in itertools.product(*self.domains):
            yield self.makePoint(*values)

    def __eq__(self, other):
        if type(other) is not Struct:
            return NotImplemented
        return self.namedDomains == other.namedDomains

    def __hash__(self):
        return hash(self.namedDomains)

    def __repr__(self):
        return f'Struct({self.domainNamed})'

class FilteredDomain(Domain):
    """A domain filtered from another by an arbitrary function."""

    def __init__(self, domain, filterFunction):
        self.domain = domain
        self.filter = filterFunction

    def uniformPoint(self):
        sample = self.domain.uniformPoint()
        if self.filter(sample):
            return sample
        else:
            raise RejectionException

    def flattenOnto(self, point, targetList):
        self.domain.flattenOnto(point, targetList)

    @cached_property
    def flattenedDimension(self):
        return self.domain.flattenedDimension

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        return self.domain.meaningOfFlatCoordinate(index,
                                                   pointName=pointName)

    def pandasIndexForFlatCoordinate(self, index):
        return self.domain.pandasIndexForFlatCoordinate(index)

    def coordinateIsNumerical(self, index):
        return self.domain.coordinateIsNumerical(index)

    def unflattenIterator(self, coords):
        return self.domain.unflattenIterator(coords)

    @cached_property
    def requiresRejection(self):
        return True

    @cached_property
    def pointsAreScalars(self):
        return self.domain.pointsAreScalars

    def __iter__(self):
        return filter(self.filter, self.domain)

    def __eq__(self, other):
        if type(other) is not FilteredDomain:
            return NotImplemented
        return (self.domain == other.domain and self.filter == other.filter)

    def __hash__(self):
        return hash((self.domain, self.filter))

    def __repr__(self):
        return f'FilteredDomain({self.domain}, {self.filter})'

### Features

class Feature:
    """A feature or list of features with unknown length.

    Args:
        domain: a `Domain` object specifying the Feature's possible values;
        distribution (optional): object specifying the distribution of values;
        lengthDomain (`Domain`): if not None, this `Feature` is actually a list
          of features, with possible lengths given by this `Domain`;
        lengthDistribution (optional): distribution over lengths;
        distanceMetric (optional): if not None, custom distance metric.

    .. testcode::

        # Feature consisting of list of 10 cars
        carDomain = Struct({
            'position': Array(Real(), [3]),
            'heading': Box((0, math.pi))
        })
        Feature(Array(carDomain, [10]))

        # Feature consisting of list of 1-10 cars
        Feature(carDomain, lengthDomain=DiscreteBox((1, 10)))
    """
    def __init__(self,
                 domain,
                 distribution=None,
                 lengthDomain=None,
                 lengthDistribution=None,
                 distanceMetric=None):
        self.domain = domain
        self.distribution = distribution
        self.lengthDomain = lengthDomain
        self.lengthDistribution = lengthDistribution
        if lengthDomain is None:
            self.minLength = self.maxLength = None
        else:
            if not isinstance(lengthDomain, DiscreteBox):
                raise RuntimeError(
                    'Feature length domain must be a DiscreteBox')
            if lengthDomain.dimension != 1:
                raise RuntimeError('Feature length domain must be an interval')
            self.minLength, self.maxLength = lengthDomain.intervals[0]
            if self.minLength < 0:
                raise RuntimeError('Feature length domain must be nonnegative')
            if (lengthDistribution is not None
                    and len(lengthDistribution.dependencies) > 0):
                raise RuntimeError(
                    'lengths of feature lists must be primitive distributions')
        self.distanceMetric = distanceMetric

    def distance(self, valueA, valueB):
        if self.distanceMetric is None:
            return self.domain.distance(valueA, valueB)
        else:
            return self.distanceMetric(valueA, valueB)

    @cached_property
    def fixedDomains(self):
        """Return the fixed-length Domains associated with this feature."""
        if not self.lengthDomain:
            return self.domain
        domains = {}
        for length in self.lengthDomain:
            length = length[0]
            domains[length] = Array(self.domain, (length,))
        return domains

    def __repr__(self):
        rep = f'Feature({self.domain}'
        if self.distribution is not None:
            rep += f', distribution={self.distribution}'
        if self.lengthDomain is not None:
            rep += f', lengthDomain={self.lengthDomain}'
        if self.lengthDistribution is not None:
            rep += f', lengthDistribution={self.lengthDistribution}'
        if self.distanceMetric is not None:
            rep += f', distanceMetric={self.distanceMetric}'
        return rep + ')'

### Feature spaces

class FeatureSpace:
    """A space consisting of named features.

    .. testcode::

        FeatureSpace({
            'weather': Feature(DiscreteBox([0, 12])),
            'egoCar': Feature(carDomain),
            'traffic': Feature(Array(carDomain, [4]))
        })
    """

    def __init__(self, features, distanceMetric=None):
        self.namedFeatures = tuple(sorted(features.items(), key=lambda i: i[0]))
        self.featureNamed = OrderedDict(self.namedFeatures)
        self.features = tuple(self.featureNamed.values())
        self.makePoint = namedtuple('SpacePoint', self.featureNamed.keys())
        self.distanceMetric = distanceMetric

    @cached_property
    def domains(self):
        """Return the domain or domains associated with this space.

        Returns a pair consisting of the Domain of all lengths of feature
        lists, plus a dict mapping each (flattened) point in that Domain to the
        corresponding Domain of other features. If the FeatureSpace has no
        feature lists, then returns (None, dom) where dom is the fixed Domain
        of all features.
        """
        fixedDomains = {}
        lengthDomains = {}
        variableDomains = {}
        for name, feature in self.namedFeatures:
            if feature.lengthDomain:
                lengthDomains[name] = feature.lengthDomain
                variableDomains[name] = feature.fixedDomains
            else:
                fixedDomains[name] = feature.domain
        if len(lengthDomains) == 0:
            return (None, Struct(fixedDomains))
        lengthDomain = Struct(lengthDomains)
        domainForLengths = {}
        for lengths in lengthDomain:
            domains = dict(fixedDomains)
            # add domains for all feature lists with nonzero lengths
            for name, length in zip(lengths._fields, lengths):
                vdomains = variableDomains[name]
                length = length[0]
                domains[name] = vdomains[length]
            # combine domains into whole domain for this assignment of lengths
            domainForLengths[lengths] = Struct(domains)
        return (lengthDomain, domainForLengths)

    def distance(self, pointA, pointB):
        distances = tuple(
            feature.distance(subptA, subptB)
            for feature, subptA, subptB in zip(self.features, pointA, pointB)
        )
        if self.distanceMetric is None:
            return np.linalg.norm(distances)
        else:
            return self.distanceMetric(distances)

    def flatten(self, point, fixedDimension=False):
        """Flatten a point in this space. See Domain.flatten.

        If fixedDimension is True, the point is flattened out as if all feature
        lists had their maximum lengths, with None as a placeholder. This means
        that all points in the space will flatten to the same length.
        """
        flattened = []
        for feature, value in zip(self.features, point):
            domain = feature.domain
            if feature.lengthDomain:
                length = len(value)
                flattened.append(length)
                fixedDomain = feature.fixedDomains[length]
                fixedDomain.flattenOnto(value, flattened)
                if fixedDimension:      # add padding to maximum length
                    sizePerElt = domain.flattenedDimension
                    needed = (feature.maxLength - length) * sizePerElt
                    for i in range(needed):
                        flattened.append(None)
            else:
                domain.flattenOnto(value, flattened)
        return tuple(flattened)

    @cached_property
    def fixedFlattenedDimension(self):
        """Length of vector returned by flatten with fixedDimension=True.

        Also an upper bound on the length of the vector returned by flatten
        by default, when fixedDimension=False."""
        dim = 0
        for feature in self.features:
            domain = feature.domain
            if feature.lengthDomain:
                dim += 1    # dimension storing length of the feature list
                dim += feature.maxLength * domain.flattenedDimension
            else:
                dim += domain.flattenedDimension
        return dim

    def meaningOfFlatCoordinate(self, index, pointName='point'):
        """Meaning of a coordinate of a flattened point in this space.

        See the corresponding function of Domain. Works only for points
        flattened with fixedDimension=True, since otherwise a given index can
        have different meaning depending on the lengths of feature lists.
        """
        assert 0 <= index < self.fixedFlattenedDimension
        for name, feature in self.namedFeatures:
            domain = feature.domain
            if feature.lengthDomain:
                if index == 0:
                    return f'len({pointName}.{name})'
                else:
                    index -= 1
                    elem = index // domain.flattenedDimension
                    if elem < feature.maxLength:
                        subIndex = index % domain.flattenedDimension
                        subPoint = f'{pointName}.{name}[{elem}]'
                        return domain.meaningOfFlatCoordinate(subIndex,
                            pointName=subPoint)
                    index -= feature.maxLength * domain.flattenedDimension
            else:
                if index < domain.flattenedDimension:
                    subPoint = f'{pointName}.{name}'
                    return domain.meaningOfFlatCoordinate(index,
                                                          pointName=subPoint)
                index -= domain.flattenedDimension
        raise RuntimeError('impossible index arithmetic')

    def pandasIndexForFlatCoordinate(self, index):
        """Pandas index of a coordinate of a flattened point in this space.

        See meaningOfFlatCoordinate, and Domain.pandasIndexForFlatCoordinate.
        """
        assert 0 <= index < self.fixedFlattenedDimension
        for name, feature in self.namedFeatures:
            domain = feature.domain
            if feature.lengthDomain:
                if index == 0:
                    return (name, 'length')
                else:
                    index -= 1
                    elem = index // domain.flattenedDimension
                    if elem < feature.maxLength:
                        subIndex = index % domain.flattenedDimension
                        panda = domain.pandasIndexForFlatCoordinate(subIndex)
                        return (name, elem) + panda
                    index -= feature.maxLength * domain.flattenedDimension
            else:
                if index < domain.flattenedDimension:
                    panda = domain.pandasIndexForFlatCoordinate(index)
                    return (name,) + panda
                index -= domain.flattenedDimension
        raise RuntimeError('impossible index arithmetic')

    def coordinateIsNumerical(self, index):
        """Whether the value of a coordinate is intrinsically numerical.

        See meaningOfFlatCoordinate, and Domain.coordinateIsNumerical.
        """
        assert 0 <= index < self.fixedFlattenedDimension
        for name, feature in self.namedFeatures:
            domain = feature.domain
            if feature.lengthDomain:
                if index == 0:
                    return True
                else:
                    index -= 1
                    elem = index // domain.flattenedDimension
                    if elem < feature.maxLength:
                        subIndex = index % domain.flattenedDimension
                        return domain.coordinateIsNumerical(subIndex)
                    index -= feature.maxLength * domain.flattenedDimension
            else:
                if index < domain.flattenedDimension:
                    return domain.coordinateIsNumerical(index)
                index -= domain.flattenedDimension
        raise RuntimeError('impossible index arithmetic')

    def unflatten(self, coords, fixedDimension=False):
        """Unflatten a tuple of coordinates to a point in this space."""
        values = []
        iterator = iter(coords)
        for feature in self.features:
            domain = feature.domain
            if feature.lengthDomain:
                length = next(iterator)
                fixedDomain = feature.fixedDomains[length]
                values.append(fixedDomain.unflattenIterator(iterator))
                if fixedDimension:      # consume padding
                    sizePerElt = domain.flattenedDimension
                    needed = (feature.maxLength - length) * sizePerElt
                    for i in range(needed):
                        next(iterator)
            else:
                values.append(domain.unflattenIterator(iterator))
        return self.makePoint(*values)

    def __repr__(self):
        rep = f'FeatureSpace({self.featureNamed}'
        if self.distanceMetric is not None:
            rep += f', distanceMetric={self.distanceMetric}'
        return rep + ')'



# This file has examples of simple control

# Example 1 : Simple rendering of a scene with a car
from verifai.simulators.car_simulator.simulator import *
from verifai.simulators.car_simulator.lane import *
from verifai.simulators.car_simulator.car_object import *
import numpy as np

def stalled_car(width=0.13):
    lanes = []
    lanes.append(straight_lane([0., -1.], [0., 1.], width))
    lanes.append(lanes[0].shifted(1))
    lanes.append(lanes[0].shifted(-1))

    # Defining a standing car
    cars = []
    car = bicycle_model(x0= np.array([0.0, 0.0, 0.0, np.pi/2]),
                        u_domain= {'omega':[-np.pi/4, np.pi/4],
                                   'acc':[-4.0, 4.0]},
                        compute_control= lambda x, u: np.array([0.0, 0.0]),
                        color='red')
    cars.append(car)

    world = simulator_world(lanes=lanes, cars=cars)

    sim = simulator()
    sim.init_simulator(world=world, task_name='Stalled Car')
    sim.run()
    sim.exit_simulation()

def constant_speed_car(width=0.13, speed=1.0):
    lanes = []
    lanes.append(straight_lane([0., -1.], [0., 1.], width))
    lanes.append(lanes[0].shifted(1))
    lanes.append(lanes[0].shifted(-1))

    # Defining a standing car
    cars = []
    car = bicycle_model(x0= np.array([0.0, 0.0, speed, np.pi/2]),
                        u_domain= {'omega':[-np.pi/4, np.pi/4],
                                   'acc':[-4.0, 4.0]},
                        compute_control= lambda x, u: np.array([0.0, 0.0]),
                        color='red')
    cars.append(car)

    world = simulator_world(lanes=lanes, cars=cars)

    sim = simulator()
    sim.init_simulator(world=world, task_name='Constant Speed')
    sim.run()
    sim.exit_simulation()

constant_speed_car()


# This file is a modification of existing CARLA code:

# Copyright (c) 2018 Intel Labs.
# authors: German Ros (german.ros@intel.com)
#
# This work is licensed under the terms of the MIT license.
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import print_function

import argparse
import collections
import datetime
import glob
import logging
import math
import os
import random
import re
import sys
import weakref

import pygame
import numpy as np

import carla
from carla import ColorConverter as cc


# ==============================================================================
# -- Global functions ----------------------------------------------------------
# ==============================================================================

def find_weather_presets():
    rgx = re.compile('.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)')
    name = lambda x: ' '.join(m.group(0) for m in rgx.finditer(x))
    presets = [x for x in dir(carla.WeatherParameters) if re.match('[A-Z].+', x)]
    return [(getattr(carla.WeatherParameters, x), name(x)) for x in presets]


def get_actor_display_name(actor, truncate=250):
    name = ' '.join(actor.type_id.replace('_', '.').title().split('.')[1:])
    return (name[:truncate-1] + u'\u2026') if len(name) > truncate else name

class Entity():
    ''' Bundles an actor in the Carla world with any corresponding data (e.g. blueprint, spawn point).'''
    def __init__(self, world, blueprint_filter,
                 spawn=None, color=None, ego=False):
        self.world = world      # Not a CARLA world, but the wrapper World class defined below
        self.blueprint_filter = blueprint_filter
        self.spawn = spawn
        self.actor = None       # Created when entity spawned.
        self.ego = ego
        self.color = color
        blueprints = self.world.world.get_blueprint_library()
        if self.blueprint_filter:
            blueprints = blueprints.filter(self.blueprint_filter)
        self.blueprint = np.random.choice(blueprints)
        if self.blueprint.has_attribute('color'):
            if self.color is None:
                self.color = random.choice(
                    self.blueprint.get_attribute('color').recommended_values)
            self.blueprint.set_attribute('color', self.color)

    def destroy(self):
        if self.actor is not None:
            self.actor.destroy()
        self.actor = None

    def create(self):
        if self.ego:
            self.blueprint.set_attribute('role_name', 'hero')
        self.actor = None
        tries = 0
        while not self.actor:
            spawn = self.spawn if self.spawn \
                else np.random.choice(self.world.map.get_spawn_points())
            self.actor = self.world.world.try_spawn_actor(self.blueprint, spawn)
            tries += 1
            if tries > 10:
                print('Unable to spawn entity at', self.spawn.location.x,
                        self.spawn.location.y, self.spawn.location.z)
                return False
        return True


class Pedestrian(Entity):
    def __init__(self, world, blueprint_filter='walker.pedestrian.*',
                 spawn=None, color=None, ego=False):
        # Pedestrian cannot be ego:
        assert not ego
        super().__init__(world, blueprint_filter,
                         spawn=spawn, color=color, ego=False)

    def create(self):
        if not super().create():
            return False
        return True

    def destroy(self):
        super().destroy()

class Prop(Entity):
    def __init__(self, world, blueprint_filter='static.prop.*',
                 spawn=None, ego=False):
        # Prop cannot be ego:
        assert not ego
        super().__init__(world, blueprint_filter=blueprint_filter,
                         spawn=spawn, ego=False)

    def create(self):
        if not super().create():
            return False
        return True

    def destroy(self):
        super().destroy()


class Vehicle(Entity):
    ''' Bundles up a CARLA Vehicle with its controller, blueprint, spawn point.'''
    def __init__(self, world, controller, control_params=None,
                 blueprint_filter='vehicle.*', spawn=None, physics=None,
                 has_collision_sensor=False, has_lane_sensor=False, color=None, ego=False):
        super().__init__(world, blueprint_filter,
                         spawn=spawn, color=color, ego=ego)
        self.ego = ego
        self.physics = physics  # CARLA PhysicsControl object.
        self.has_collision_sensor = has_collision_sensor # Should be True if ego.
        self.collision_sensor = None
        self.has_lane_sensor = has_lane_sensor
        self.lane_sensor = None
        self.controller = controller
        self.control_params = control_params
        self.control_agent = None

    def destroy(self):
        super().destroy()
        if self.collision_sensor is not None:
            self.collision_sensor.sensor.destroy()
        self.collision_sensor = None
        if self.lane_sensor is not None:
            self.lane_sensor.sensor.destroy()
        self.lane_sensor = None

    def create(self):
        if not super().create():
            return False
        if self.ego:
            # Keep same camera config if the camera manager exists in World.
            # cam_index = self.world.camera_manager._index \
            #     if self.world.camera_manager is not None else 0
            # cam_pos_index = self.world.camera_manager._transform_index \
            #     if self.world.camera_manager is not None else 0
            cam_index = 0
            cam_pos_index = 0
            self.world.camera_manager = CameraManager(self.actor, self.world.hud)
            self.world.camera_manager._transform_index = cam_pos_index
            self.world.camera_manager.set_sensor(cam_index, notify=False)
            self.world.camera_manager.set_transform(self.world.cam_transform)
            actor_type = get_actor_display_name(self.actor)
            self.world.hud.notification(actor_type)

        if self.has_collision_sensor:
            self.collision_sensor = CollisionSensor(self.actor,
                                                    hud=self.world.hud if self.ego else None)

        if self.has_lane_sensor:
            self.lane_sensor = LaneInvasionSensor(self.actor,
                                                  hud=self.world.hud if self.ego else None)

        if self.physics:
            self.actor.apply_physics_control(self.physics)

        if self.control_params:
            self.control_actor = self.controller(self.actor, self.control_params)
        else:
            self.control_actor = self.controller(self.actor)

        return True

    def get_control(self):
        control = self.control_actor.run_step()
        return control


# ==============================================================================
# -- World ---------------------------------------------------------------
# ==============================================================================

class World(object):
    def __init__(self, carla_world, hud, cam_transform=0):
        self.world = carla_world
        self.map = self.world.get_map()
        self.hud = hud
        self.entities = []
        self.ego = None
        self.camera_manager = None
        self._weather_presets = find_weather_presets()
        self._weather_index = 0
        self.restart()
        self.world.on_tick(hud.on_world_tick)
        self.cam_transform = cam_transform
        self.ego_images = []

    def add_vehicle(self, controller, control_params=None, 
                    blueprint_filter='vehicle.*', color=None,
                    spawn=None, physics=None, has_collision_sensor=False,
                    has_lane_sensor=False, ego=False):
        '''
        Add vehicle to vehicles list with controller CONTROLLER, blueprint 
        BLUEPRINT and spawn point SPAWN. If EGO, this will be the ego vehicle 
        attached to camera manager, HUD, sensors. BLUEPRINT and SPAWN default 
        to being chosen randomly.
        '''
        vehicle = Vehicle(self, controller, control_params=control_params,
                          blueprint_filter=blueprint_filter, color=color, 
                          spawn=spawn, physics=physics,
                          has_collision_sensor=has_collision_sensor,
                          has_lane_sensor=has_lane_sensor, ego=ego)
        self.entities.append(vehicle)
        if ego:
            self.ego = vehicle
        return vehicle

    def add_pedestrian(self, blueprint_filter='walker.pedestrian.*',
                       spawn=None, color=None, ego=False):
        ped = Pedestrian(self, blueprint_filter=blueprint_filter,
                         spawn=spawn, color=color, ego=ego)
        self.entities.append(ped)
        return ped

    def add_prop(self, blueprint_filter='static.prop.*',
                       spawn=None, ego=False):
        prop = Prop(self, blueprint_filter=blueprint_filter,
                         spawn=spawn, ego=ego)
        self.entities.append(prop)
        return prop

    def restart(self):
        # Destroy all entities in world first.
        self.destroy()

        # Spawn all entities.
        for e in self.entities:
            e.create()

    def get_control_cmds(self):
        return [carla.command.ApplyVehicleControl(v.actor.id, v.get_control())
                for v in self.entities if isinstance(v, Vehicle) and v.actor is not None]

    def next_weather(self, reverse=False):
        self._weather_index += -1 if reverse else 1
        self._weather_index %= len(self._weather_presets)
        preset = self._weather_presets[self._weather_index]
        self.hud.notification('Weather: %s' % preset[1])
        self.world.set_weather(preset[0])

    def tick(self, clock):
        self.hud.tick(self, clock)

    def render(self, display):
        self.camera_manager.render(display)
        self.hud.render(display)

    def destroy(self): 
        for e in self.entities:
            if e is not None:
                e.destroy()

        if self.camera_manager is not None:
            self.camera_manager.sensor.destroy()
            self.camera_manager.images = []


# ==============================================================================
# -- HUD -----------------------------------------------------------------
# ==============================================================================

class HUD(object):
    def __init__(self, width, height, show_hud=False):
        self.dim = (width, height)
        self._show_info = show_hud
        if self._show_info:
            font = pygame.font.Font(pygame.font.get_default_font(), 20)
            fonts = [x for x in pygame.font.get_fonts() if 'mono' in x]
            default_font = 'ubuntumono'
            mono = default_font if default_font in fonts else fonts[0]
            mono = pygame.font.match_font(mono)
            self._font_mono = pygame.font.Font(mono, 14)
            self._notifications = FadingText(font, (width, 40), (0, height - 40))
        self.server_fps = 0
        self.frame = 0
        self.simulation_time = 0
        self._info_text = []
        self._server_clock = pygame.time.Clock()

    def on_world_tick(self, timestamp):
        self._server_clock.tick()
        self.server_fps = self._server_clock.get_fps()
        self.frame = timestamp.frame
        self.simulation_time = timestamp.elapsed_seconds

    def tick(self, world, clock):
        if not self._show_info:
            return
        if world.ego.actor is None:
            # Ego not spawned yet
            return
        t = world.ego.actor.get_transform()
        v = world.ego.actor.get_velocity()
        c = world.ego.actor.get_control()
        heading = 'N' if abs(t.rotation.yaw) < 89.5 else ''
        heading += 'S' if abs(t.rotation.yaw) > 90.5 else ''
        heading += 'E' if 179.5 > t.rotation.yaw > 0.5 else ''
        heading += 'W' if -0.5 > t.rotation.yaw > -179.5 else ''
        colhist = world.ego.collision_sensor.get_collision_history()
        collision = [colhist[x + self.frame - 200] for x in range(0, 200)]
        max_col = max(1.0, max(collision))
        collision = [x / max_col for x in collision]
        vehicles = world.world.get_actors().filter('vehicle.*')
        self._info_text = [
            'Server:  % 16d FPS' % self.server_fps,
            '',
            'Vehicle: % 20s' % get_actor_display_name(world.ego.actor, truncate=20),
            'Map:     % 20s' % world.map.name,
            'Simulation time: % 12s' % datetime.timedelta(seconds=int(self.simulation_time)),
            '',
            'Speed:   % 15.0f km/h' % (3.6 * math.sqrt(v.x**2 + v.y**2 + v.z**2)),
            u'Heading:% 16.0f\N{DEGREE SIGN} % 2s' % (t.rotation.yaw, heading),
            'Location:% 20s' % ('(% 5.1f, % 5.1f)' % (t.location.x, t.location.y)),
            'Height:  % 18.0f m' % t.location.z,
            '',
            ('Throttle:', c.throttle, 0.0, 1.0),
            ('Steer:', c.steer, -1.0, 1.0),
            ('Brake:', c.brake, 0.0, 1.0),
            ('Reverse:', c.reverse),
            ('Hand brake:', c.hand_brake),
            ('Manual:', c.manual_gear_shift),
            'Gear:        %s' % {-1: 'R', 0: 'N'}.get(c.gear, c.gear),
            '',
            'Collision:',
            collision,
            '',
            'Number of vehicles: % 8d' % len(vehicles)
        ]
        if len(vehicles) > 1:
            self._info_text += ['Nearby vehicles:']
            distance = lambda l: math.sqrt((l.x - t.location.x)**2 + (l.y - t.location.y)**2 + (l.z - t.location.z)**2)
            vehicles = [(distance(x.get_location()), x)
                        for x in vehicles if x.id != world.ego.actor.id]
            for d, vehicle in sorted(vehicles):
                if d > 200.0:
                    break
                vehicle_type = get_actor_display_name(vehicle, truncate=22)
                self._info_text.append('% 4dm %s' % (d, vehicle_type))
        # self._notifications.tick(world, clock)
        pass

    def toggle_info(self):
        self._show_info = not self._show_info

    def notification(self, text, seconds=2.0):        
        # self._notifications.set_text(text, seconds=seconds)
        pass

    def error(self, text):
        # self._notifications.set_text('Error: %s' % text, (255, 0, 0))
        pass

    def render(self, display):
        if self._show_info:
            info_surface = pygame.Surface((220, self.dim[1]))
            info_surface.set_alpha(100)
            display.blit(info_surface, (0, 0))
            v_offset = 4
            bar_h_offset = 100
            bar_width = 106
            for item in self._info_text:
                if v_offset + 18 > self.dim[1]:
                    break
                if isinstance(item, list):
                    if len(item) > 1:
                        points = [(x + 8, v_offset + 8 + (1.0 - y) * 30) for x, y in enumerate(item)]
                        pygame.draw.lines(display, (255, 136, 0), False, points, 2)
                    item = None
                    v_offset += 18
                elif isinstance(item, tuple):
                    if isinstance(item[1], bool):
                        rect = pygame.Rect((bar_h_offset, v_offset + 8), (6, 6))
                        pygame.draw.rect(display, (255, 255, 255), rect, 0 if item[1] else 1)
                    else:
                        rect_border = pygame.Rect((bar_h_offset, v_offset + 8), (bar_width, 6))
                        pygame.draw.rect(display, (255, 255, 255), rect_border, 1)
                        f = (item[1] - item[2]) / (item[3] - item[2])
                        if item[2] < 0.0:
                            rect = pygame.Rect((bar_h_offset + f * (bar_width - 6), v_offset + 8), (6, 6))
                        else:
                            rect = pygame.Rect((bar_h_offset, v_offset + 8), (f * bar_width, 6))
                        pygame.draw.rect(display, (255, 255, 255), rect)
                    item = item[0]
                if item: # At this point has to be a str.
                    surface = self._font_mono.render(item, True, (255, 255, 255))
                    display.blit(surface, (8, v_offset))
                v_offset += 18
        # self._notifications.render(display)
        pass


# ==============================================================================
# -- FadingText ----------------------------------------------------------------
# ==============================================================================

class FadingText(object):
    def __init__(self, font, dim, pos):
        self.font = font
        self.dim = dim
        self.pos = pos
        self.seconds_left = 0
        self.surface = pygame.Surface(self.dim)

    def set_text(self, text, color=(255, 255, 255), seconds=2.0):
        text_texture = self.font.render(text, True, color)
        self.surface = pygame.Surface(self.dim)
        self.seconds_left = seconds
        self.surface.fill((0, 0, 0, 0))
        self.surface.blit(text_texture, (10, 11))

    def tick(self, _, clock):
        delta_seconds = 1e-3 * clock.get_time()
        self.seconds_left = max(0.0, self.seconds_left - delta_seconds)
        self.surface.set_alpha(500.0 * self.seconds_left)

    def render(self, display):
        display.blit(self.surface, self.pos)


# ==============================================================================
# -- CollisionSensor -----------------------------------------------------------
# ==============================================================================


class CollisionSensor(object):
    def __init__(self, parent_actor, hud=None):
        self.sensor = None
        self._history = []
        self._parent = parent_actor
        self._parent_mass = parent_actor.get_physics_control().mass
        self._hud = hud
        world = self._parent.get_world()
        bp = world.get_blueprint_library().find('sensor.other.collision')
        self.sensor = world.spawn_actor(bp, carla.Transform(), attach_to=self._parent)
        # We need to pass the lambda a weak reference to self to avoid circular
        # reference.
        weak_self = weakref.ref(self)
        self.sensor.listen(lambda event: CollisionSensor._on_collision(weak_self, event))

    def get_collision_speeds(self):
        '''Convert collision intensities from momentem (kg*m/s) to speed (km/h).'''
        return [(c[0], 3.6 * c[1] / self._parent_mass)
                         for c in self._history]

    def get_collision_history(self):
        history = collections.defaultdict(int)
        for frame, intensity in self._history:
            history[frame] += intensity
        return history

    @staticmethod
    def _on_collision(weak_self, event):
        self = weak_self()
        if not self:
            return
        actor_type = get_actor_display_name(event.other_actor)
        if self._hud:
            self._hud.notification('Collision with %r, id = %d'
                                   % (actor_type, event.other_actor.id))
        impulse = event.normal_impulse
        intensity = math.sqrt(impulse.x ** 2 + impulse.y ** 2 + impulse.z ** 2)
        self._history.append((event.frame, intensity))
        if len(self._history) > 4000:
            self._history.pop(0)


# ==============================================================================
# -- LaneInvasionSensor --------------------------------------------------------
# ==============================================================================

class LaneInvasionSensor(object):
    def __init__(self, parent_actor, hud=None):
        self.sensor = None
        self._parent = parent_actor
        self._hud = hud
        self._history = []
        world = self._parent.get_world()
        bp = world.get_blueprint_library().find('sensor.other.lane_detector')
        self.sensor = world.spawn_actor(bp, carla.Transform(), attach_to=self._parent)
        # We need to pass the lambda a weak reference to self to avoid circular
        # reference.
        weak_self = weakref.ref(self)
        self.sensor.listen(lambda event: LaneInvasionSensor._on_invasion(weak_self, event))

    @staticmethod
    def _on_invasion(weak_self, event):
        self = weak_self()
        if not self:
            return
        text = ['%r' % str(x).split()[-1] for x in set(event.crossed_lane_markings)]
        if self._hud:
            self._hud.notification('Crossed line %s' % ' and '.join(text))
        self._history.append((event.frame, 1))


# ==============================================================================
# -- CameraManager -------------------------------------------------------------
# ==============================================================================

class CameraManager(object):
    def __init__(self, parent_actor, hud):
        self.sensor = None
        self._surface = None
        self._parent = parent_actor
        self._hud = hud
        self._recording = False
        self.images = []
        self._camera_transforms = [
            carla.Transform(carla.Location(x=-5.5, z=2.8), carla.Rotation(pitch=-15)),
            carla.Transform(carla.Location(x=1.6, z=1.7))]
        self._transform_index = 1
        self._sensors = [
            ['sensor.camera.rgb', cc.Raw, 'Camera RGB'],
            ['sensor.camera.depth', cc.Raw, 'Camera Depth (Raw)'],
            ['sensor.camera.depth', cc.Depth, 'Camera Depth (Gray Scale)'],
            ['sensor.camera.depth', cc.LogarithmicDepth, 'Camera Depth (Logarithmic Gray Scale)'],
            ['sensor.camera.semantic_segmentation', cc.Raw, 'Camera Semantic Segmentation (Raw)'],
            ['sensor.camera.semantic_segmentation', cc.CityScapesPalette,
             'Camera Semantic Segmentation (CityScapes Palette)'],
            ['sensor.lidar.ray_cast', None, 'Lidar (Ray-Cast)']]
        world = self._parent.get_world()
        bp_library = world.get_blueprint_library()
        for item in self._sensors:
            bp = bp_library.find(item[0])
            if item[0].startswith('sensor.camera'):
                bp.set_attribute('image_size_x', str(hud.dim[0]))
                bp.set_attribute('image_size_y', str(hud.dim[1]))
            item.append(bp)
        self._index = None

    def toggle_camera(self):
        set_transform((self._transform_index + 1) % len(self._camera_transforms))

    def set_transform(self, idx):
        self._transform_index = idx
        self.sensor.set_transform(self._camera_transforms[self._transform_index])

    def set_sensor(self, index, notify=True):
        index = index % len(self._sensors)
        needs_respawn = True if self._index is None \
            else self._sensors[index][0] != self._sensors[self._index][0]
        if needs_respawn:
            if self.sensor is not None:
                self.sensor.destroy()
                self._surface = None
            self.sensor = self._parent.get_world().spawn_actor(
                self._sensors[index][-1],
                self._camera_transforms[self._transform_index],
                attach_to=self._parent)
            # We need to pass the lambda a weak reference to self to avoid
            # circular reference.
            weak_self = weakref.ref(self)
            self.sensor.listen(lambda image: CameraManager._parse_image(weak_self, image))
        if notify:
            self._hud.notification(self._sensors[index][2])
        self._index = index

    def next_sensor(self):
        self.set_sensor(self._index + 1)

    def toggle_recording(self):
        self._recording = not self._recording
        self._hud.notification('Recording %s' % ('On' if self._recording else 'Off'))

    def render(self, display):
        if self._surface is not None:
            display.blit(self._surface, (0, 0))

    @staticmethod
    def _parse_image(weak_self, image):
        self = weak_self()
        if not self:
            return
        if self._sensors[self._index][0].startswith('sensor.lidar'):
            points = np.frombuffer(image.raw_data, dtype=np.dtype('f4'))
            points = np.reshape(points, (int(points.shape[0] / 3), 3))
            lidar_data = np.array(points[:, :2])
            lidar_data *= min(self._hud.dim) / 100.0
            lidar_data += (0.5 * self._hud.dim[0], 0.5 * self._hud.dim[1])
            lidar_data = np.fabs(lidar_data)
            lidar_data = lidar_data.astype(np.int32)
            lidar_data = np.reshape(lidar_data, (-1, 2))
            lidar_img_size = (self._hud.dim[0], self._hud.dim[1], 3)
            lidar_img = np.zeros(lidar_img_size)
            lidar_img[tuple(lidar_data.T)] = (255, 255, 255)
            self._surface = pygame.surfarray.make_surface(lidar_img)
        else:
            image.convert(self._sensors[self._index][1])
            array = np.frombuffer(image.raw_data, dtype=np.dtype("uint8"))
            array = np.reshape(array, (image.height, image.width, 4))
            array = array[:, :, :3]
            array = array[:, :, ::-1]
            self._surface = pygame.surfarray.make_surface(array.swapaxes(0, 1))
        if self._recording:
            image.save_to_disk('_out/%08d' % image.frame)
        self.images.append(image)



"""Bayesian Optimization sampler : Defined only for continuous domains.
For discrete inputs define another sampler"""

from verifai.samplers.domain_sampler import BoxSampler
import numpy as np


class BayesOptSampler(BoxSampler):
    def __init__(self, domain, BO_params):
        try:
            import GPyOpt
        except ModuleNotFoundError:
            raise RuntimeError(
                'BayesOptSampler requires GPyOpt to be installed')

        super().__init__(domain)
        self.init_num = BO_params.init_num
        if self.init_num < 1:
            raise RuntimeError(
                'init_num for BayesOptSampler must be at least 1')
        self.bounds = []
        for i in range(self.dimension):
            self.bounds.append({'name':'x_'+str(i), 'type': 'continuous',
                                'domain': (0,1)})
        self.X = np.empty((0, self.dimension))
        self.Y = np.empty((0, 1))

    def getVector(self, feedback=None):
        import GPyOpt   # do this here to avoid slow import when unused

        if len(self.X) < self.init_num:
            # Do random sampling
            sample = np.random.uniform(0, 1, self.dimension)
        else:
            BO = GPyOpt.methods.BayesianOptimization(
                f=None, batch_size=1,
                domain=self.bounds, X=self.X, Y=self.Y, normalize_Y=False)
            sample = BO.suggest_next_locations()[0]
        return tuple(sample), None

    def updateVector(self, vector, info, rho):
        self.X = np.vstack((self.X, np.atleast_2d(vector)))
        self.Y = np.vstack((self.Y, np.atleast_2d(rho)))



from verifai.simulators.carla.client_carla import *
from verifai.simulators.carla.carla_world import *
from verifai.simulators.carla.carla_task import *

import numpy as np
from dotmap import DotMap
import time
import carla
from carla import Transform, Rotation, Location

from verifai.simulators.carla.agents.brake_agent import *
from verifai.simulators.carla.agents.pid_agent import *

WEATHER_PARAMS = ['cloudiness', 'precipitation', 'precipitation_deposits',
                  'wind_intensity', 'sun_azimuth_angle', 'sun_altitude_angle']

AGENTS = {'BrakeAgent': BrakeAgent, 'PIDAgent': PIDAgent}

class scenic_sampler_task(carla_task):
    def __init__(self,
                 n_sim_steps=100,
                 display_dim=(1280,720),
                 carla_host='127.0.0.1',
                 carla_port=2000,
                 carla_timeout=4.0,
                 world_map='Town01',
                 cam_transform=0):
        super().__init__(
            n_sim_steps=n_sim_steps,
            display_dim=display_dim,
            carla_host=carla_host,
            carla_port=carla_port,
            carla_timeout=carla_timeout,
            world_map=world_map,
            cam_transform=cam_transform
        )

    def snap_to_ground(self, location):
        '''Mutates @location to have the same z-coordinate as the nearest waypoint.'''
        waypoint = self.world.world.get_map().get_waypoint(location)
        location.z = waypoint.transform.location.z + 1
        return location

    def use_sample(self, sample):
        print('USE_SAMPLE Sample:', sample)
        weather = carla.WeatherParameters(**{k: sample.params._asdict()[k] for k in WEATHER_PARAMS})
        self.world.world.set_weather(weather)
        for obj in sample.objects:
            spawn = Transform(self.snap_to_ground(Location(x=obj.position[0],
                                                           y=-obj.position[1], z=1)),
                              Rotation(yaw=-obj.heading * 180 / math.pi - 90))
            attrs = dict()
            if 'color' in obj._fields:
                color = str(int(obj.color.r * 255)) + ',' \
                    + str(int(obj.color.g * 255)) + ',' + str(int(obj.color.b * 255))
                attrs['color'] = color
            if 'blueprint' in obj._fields:
                attrs['blueprint_filter'] = obj.blueprint
            agent = BrakeAgent
            if 'agent' in obj._fields:
                agent = AGENTS[obj.agent]
            if obj.type in ['Vehicle', 'Car', 'Truck', 'Bicycle', 'Motorcycle']:
                self.world.add_vehicle(agent,
                                       spawn=spawn,
                                       has_collision_sensor=False,
                                       has_lane_sensor=False,
                                       ego=obj is sample.objects[0],
                                       **attrs)
            elif obj.type == 'Pedestrian':
                self.world.add_pedestrian(spawn=spawn, **attrs)
            elif obj.type in ['Prop', 'Trash', 'Cone']:
                self.world.add_prop(spawn=spawn, **attrs)
            else:
                print('Unsupported object type:', obj.type)

    def trajectory_definition(self):
        # To save image each iteration:
        # self.world.camera_manager.images[-1].save_to_disk(str(int(time.time())))
        return {}
